# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
import os
import shutil
import subprocess  # nosec
import sys
import xml.etree.ElementTree as ET  # nosec

import defusedxml.minidom

defusedxml.defuse_stdlib()

import warnings

from defusedxml.ElementTree import ParseError
from defusedxml.minidom import parseString

from pyaedt.generic.general_methods import read_toml
from pyaedt.generic.settings import is_linux
import pyaedt.workflows
import pyaedt.workflows.templates


def add_automation_tab(
    name,
    lib_dir,
    icon_file=None,
    product="Project",
    template="Run PyAEDT Toolkit Script",
    overwrite=False,
    panel="Panel_PyAEDT_Extensions",
):
    """Add an automation tab in AEDT.

    Parameters
    ----------
    name : str
        Toolkit name.
    lib_dir : str
        Path to the library directory.
    icon_file : str
        Full path to the icon file. The default is the PyAnsys icon.
    product : str, optional
        Product directory to install the toolkit.
    template : str, optional
        Script template name to use
    overwrite : bool, optional
        Whether to overwrite the existing automation tab. The default is ``False``, in
        which case is adding new tabs to the existing ones.
    panel : str, optional
        Panel name. The default is ``"Panel_PyAEDT_Extensions"``.

    Returns
    -------
    str
        Automation tab path.

    """

    product = __tab_map(product)

    toolkit_name = name
    if "/" in name:
        toolkit_name = name.replace("/", "_")

    tab_config_file_path = os.path.join(lib_dir, product, "TabConfig.xml")
    if not os.path.isfile(tab_config_file_path) or overwrite:
        root = ET.Element("TabConfig")
    else:
        try:
            tree = ET.parse(tab_config_file_path)  # nosec
        except ParseError as e:  # pragma: no cover
            warnings.warn("Unable to parse %s\nError received = %s" % (tab_config_file_path, str(e)))
            return
        root = tree.getroot()

    panels = root.findall("./panel")
    if panels:
        panel_names = [panel_element.attrib["label"] for panel_element in panels]
        if panel in panel_names:
            # Remove previously existing PyAEDT panel and update with newer one.
            panel_element = [panel_element for panel_element in panels if panel_element.attrib["label"] == panel][0]
        else:
            panel_element = ET.SubElement(root, "panel", label=panel)
    else:
        panel_element = ET.SubElement(root, "panel", label=panel)

    buttons = panel_element.findall("./button")
    if buttons:  # pragma: no cover
        button_names = [button.attrib["label"] for button in buttons]
        if name in button_names:
            # Remove previously existing PyAEDT panel and update with newer one.
            b = [button for button in buttons if button.attrib["label"] == name][0]
            panel_element.remove(b)

    if not icon_file:
        icon_file = os.path.join(os.path.dirname(pyaedt.workflows.__file__), "images", "large", "pyansys.png")

    file_name = os.path.basename(icon_file)

    dest_dir = os.path.normpath(os.path.join(lib_dir, product, toolkit_name, "images", "large"))
    dest_file = os.path.normpath(os.path.join(dest_dir, file_name))
    os.makedirs(os.path.dirname(dest_dir), exist_ok=True)
    if not os.path.exists(dest_dir):
        os.makedirs(dest_dir)
    shutil.copy(icon_file, dest_file)

    relative_image_path = os.path.relpath(dest_file, os.path.join(lib_dir, product))

    ET.SubElement(
        panel_element,
        "button",
        label=name,
        isLarge="1",
        image=relative_image_path,
        script="{}/{}".format(toolkit_name, template),
    )

    # Backup any existing file if present
    if os.path.isfile(tab_config_file_path):
        shutil.copy(tab_config_file_path, tab_config_file_path + ".orig")

    create_xml_tab(root, tab_config_file_path)
    return tab_config_file_path


def create_xml_tab(root, output_file):
    """Write the XML file to create the automation tab.

    Parameters
    ----------
    root : :class:xml.etree.ElementTree
        Root element of the main panel.
    output_file : str
        Full name of the file to save the XML tab.
    """
    lines = [line for line in parseString(ET.tostring(root)).toprettyxml(indent=" " * 4).split("\n") if line.strip()]
    xml_str = "\n".join(lines)

    with open(output_file, "w") as f:
        f.write(xml_str)


def remove_xml_tab(toolkit_dir, name, panel="Panel_PyAEDT_Extensions"):
    """Remove a toolkit configuration file."""
    tab_config_file_path = os.path.join(toolkit_dir, "TabConfig.xml")
    if not os.path.isfile(tab_config_file_path):  # pragma: no cover
        return True
    try:
        tree = ET.parse(tab_config_file_path)  # nosec
    except ParseError as e:  # pragma: no cover
        warnings.warn("Unable to parse %s\nError received = %s" % (tab_config_file_path, str(e)))
        return
    root = tree.getroot()

    panels = root.findall("./panel")
    if panels:
        panel_names = [panel_element.attrib["label"] for panel_element in panels]
        if panel in panel_names:
            # Remove previously existing PyAEDT panel and update with newer one.
            panel_element = [panel_element for panel_element in panels if panel_element.attrib["label"] == panel][0]
        else:  # pragma: no cover
            panel_element = ET.SubElement(root, "panel", label=panel)
    else:  # pragma: no cover
        panel_element = ET.SubElement(root, "panel", label=panel)

    buttons = panel_element.findall("./button")
    if buttons:
        button_names = [button.attrib["label"] for button in buttons]
        if name in button_names:
            # Remove previously existing PyAEDT panel and update with newer one.
            b = [button for button in buttons if button.attrib["label"] == name][0]
            panel_element.remove(b)

    create_xml_tab(root, tab_config_file_path)


def available_toolkits():
    product_list = [
        "Circuit",
        "EMIT",
        "HFSS",
        "HFSS3DLayout",
        "Icepak",
        "Maxwell2D",
        "Maxwell3D",
        "Mechanical",
        "Project",
        "Q2D",
        "Q3D",
        "TwinBuilder",
    ]

    product_toolkits = {}
    for product in product_list:
        toml_file = os.path.join(os.path.dirname(__file__), product.lower(), "toolkits_catalog.toml")
        if os.path.isfile(toml_file):
            toolkits_catalog = read_toml(toml_file)
            product_toolkits[product] = toolkits_catalog
    return product_toolkits


def add_script_to_menu(
    name,
    script_file,
    template_file="run_pyaedt_toolkit_script",
    icon_file=None,
    product="Project",
    copy_to_personal_lib=True,
    executable_interpreter=None,
    panel="Panel_PyAEDT_Extensions",
    personal_lib=None,
    aedt_version="",
):
    """Add a script to the ribbon menu.

    .. note::
       This method is available in AEDT 2023 R2 and later. PyAEDT must be installed
       in AEDT to allow this method to run. For more information, see `Installation
       <https://aedt.docs.pyansys.com/version/stable/Getting_started/Installation.html>`_.

    Parameters
    ----------
    name : str
        Name of the toolkit to appear in AEDT.
    script_file : str
        Full path to the script file. The script will be moved to Personal Lib.
    template_file : str
        Script template name to use. The default is ``"run_pyaedt_toolkit_script"``.
    icon_file : str, optional
        Full path to the icon (a 30x30 pixel PNG file) to add to the UI.
        The default is ``None``.
    product : str, optional
        Product to which the toolkit applies. The default is ``"Project"``, in which case
        it applies to all designs. You can also specify a product, such as ``"HFSS"``.
    copy_to_personal_lib : bool, optional
        Whether to copy the script to Personal Lib or link the original script. Default is ``True``.
    executable_interpreter : str, optional
        Executable python path. The default is the one current interpreter.
    panel : str, optional
        Panel name. The default is ``"Panel_PyAEDT_Extensions"``.
    personal_lib : str, optional
    aedt_version : str, optional

    Returns
    -------
    bool

    """
    logger = logging.getLogger("Global")
    if not personal_lib or not aedt_version:
        from pyaedt.generic.desktop_sessions import _desktop_sessions

        if not _desktop_sessions:  # pragma: no cover
            logger.error("Personallib or AEDT version is not provided and there is no available desktop session.")
            return False
        d = list(_desktop_sessions.values())[0]
        personal_lib = d.personallib
        aedt_version = d.aedt_version_id
    if script_file and not os.path.exists(script_file):  # pragma: no cover
        logger.error("Script does not exists.")
        return False
    toolkit_dir = os.path.join(personal_lib, "Toolkits")
    tool_map = __tab_map(product)
    file_name = name
    if "/" in file_name:  # pragma: no cover
        file_name = file_name.replace("/", "_")
    tool_dir = os.path.join(toolkit_dir, tool_map, file_name)
    lib_dir = os.path.join(tool_dir, "Lib")
    toolkit_rel_lib_dir = os.path.relpath(lib_dir, tool_dir)
    if is_linux and aedt_version <= "2023.1":  # pragma: no cover
        toolkit_rel_lib_dir = os.path.join("Lib", file_name)
        lib_dir = os.path.join(toolkit_dir, toolkit_rel_lib_dir)
        toolkit_rel_lib_dir = "../../" + toolkit_rel_lib_dir
    os.makedirs(lib_dir, exist_ok=True)
    os.makedirs(tool_dir, exist_ok=True)
    dest_script_path = None
    if script_file and copy_to_personal_lib:
        dest_script_path = os.path.join(lib_dir, os.path.split(script_file)[-1])
        shutil.copy2(script_file, dest_script_path)

    version_agnostic = True
    if aedt_version[2:6].replace(".", "") in sys.executable:  # pragma: no cover
        executable_version_agnostic = sys.executable.replace(aedt_version[2:6].replace(".", ""), "%s")
        version_agnostic = False
    else:
        executable_version_agnostic = sys.executable

    if executable_interpreter:  # pragma: no cover
        version_agnostic = True
        executable_version_agnostic = executable_interpreter

    templates_dir = os.path.dirname(pyaedt.workflows.templates.__file__)

    ipython_executable = executable_version_agnostic.replace("python" + __exe(), "ipython" + __exe())
    jupyter_executable = executable_version_agnostic.replace("python" + __exe(), "jupyter" + __exe())

    with open(os.path.join(templates_dir, template_file + ".py_build"), "r") as build_file:
        with open(os.path.join(tool_dir, template_file + ".py"), "w") as out_file:
            build_file_data = build_file.read()
            build_file_data = build_file_data.replace("##TOOLKIT_REL_LIB_DIR##", toolkit_rel_lib_dir)
            build_file_data = build_file_data.replace("##IPYTHON_EXE##", ipython_executable)
            build_file_data = build_file_data.replace("##PYTHON_EXE##", executable_version_agnostic)
            build_file_data = build_file_data.replace("##JUPYTER_EXE##", jupyter_executable)
            build_file_data = build_file_data.replace("##TOOLKIT_NAME##", file_name)
            if dest_script_path:
                build_file_data = build_file_data.replace("##PYTHON_SCRIPT##", dest_script_path)

            if version_agnostic:
                build_file_data = build_file_data.replace(" % version", "")
            out_file.write(build_file_data)

    add_automation_tab(name, toolkit_dir, icon_file=icon_file, product=product, template=template_file, panel=panel)
    logger.info("{} installed".format(name))
    return True


def __tab_map(product):  # pragma: no cover
    """Map exceptions in AEDT applications."""
    if product.lower() == "hfss3dlayout":
        return "HFSS3DLayoutDesign"
    elif product.lower() == "circuit":
        return "CircuitDesign"
    elif product.lower() == "q2d":
        return "2DExtractor"
    elif product.lower() == "q3d":
        return "Q3DExtractor"
    elif product.lower() == "simplorer":
        return "TwinBuilder"
    else:
        return product


def add_custom_toolkit(desktop_object, toolkit_name, wheel_toolkit=None, install=True):  # pragma: no cover
    """Add toolkit to AEDT Automation Tab.

    Parameters
    ----------
    desktop_object : :class:pyaedt.desktop.Desktop
        Desktop object.
    toolkit_name : str
        Name of toolkit to add.
    wheel_toolkit : str
        Wheelhouse path.
    install : bool, optional
        Whether to install the toolkit.

    Returns
    -------
    bool
    """
    toolkits = available_toolkits()
    toolkit_info = None
    product_name = None
    for product in toolkits:
        if toolkit_name in toolkits[product]:
            toolkit_info = toolkits[product][toolkit_name]
            product_name = product
            break
    if not toolkit_info:
        desktop_object.logger.error("Toolkit does not exist.")
        return False

    # Set Python version based on AEDT version
    python_version = "3.10" if desktop_object.aedt_version_id > "2023.1" else "3.7"
    python_version_new = python_version.replace(".", "_")
    if not is_linux:
        base_venv = os.path.normpath(
            os.path.join(
                desktop_object.install_path,
                "commonfiles",
                "CPython",
                python_version_new,
                "winx64",
                "Release",
                "python",
                "python.exe",
            )
        )
    else:
        base_venv = os.path.normpath(
            os.path.join(
                desktop_object.install_path,
                "commonfiles",
                "CPython",
                python_version_new,
                "linx64",
                "Release",
                "python",
                "runpython",
            )
        )

    def run_command(command):
        try:
            if is_linux:  # pragma: no cover
                process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)  # nosec
            else:
                process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)  # nosec
            _, stderr = process.communicate()
            ret_code = process.returncode
            if ret_code != 0:
                print("Error occurred:", stderr.decode("utf-8"))
            return ret_code
        except Exception as e:
            print("Exception occurred:", str(e))
            return 1  # Return non-zero exit code for indicating an error

    version = desktop_object.odesktop.GetVersion()[2:6].replace(".", "")

    if not is_linux:
        venv_dir = os.path.join(os.environ["APPDATA"], ".pyaedt_env", "toolkits_{}".format(python_version_new))
        python_exe = os.path.join(venv_dir, "Scripts", "python.exe")
        pip_exe = os.path.join(venv_dir, "Scripts", "pip.exe")
        package_dir = os.path.join(venv_dir, "Lib")
    else:
        venv_dir = os.path.join(os.environ["HOME"], ".pyaedt_env", "toolkits_{}".format(python_version_new))
        python_exe = os.path.join(venv_dir, "bin", "python")
        pip_exe = os.path.join(venv_dir, "bin", "pip")
        package_dir = os.path.join(venv_dir, "lib")
        edt_root = os.path.normpath(desktop_object.odesktop.GetExeDir())
        os.environ["ANSYSEM_ROOT{}".format(version)] = edt_root
        ld_library_path_dirs_to_add = [
            "{}/commonfiles/CPython/{}/linx64/Release/python/lib".format(edt_root, python_version_new),
            "{}/common/mono/Linux64/lib64".format(edt_root),
            "{}".format(edt_root),
        ]
        if version < "232":
            ld_library_path_dirs_to_add.append("{}/Delcross".format(edt_root))
        os.environ["LD_LIBRARY_PATH"] = ":".join(ld_library_path_dirs_to_add) + ":" + os.getenv("LD_LIBRARY_PATH", "")

    # Create virtual environment

    if not os.path.exists(venv_dir):
        desktop_object.logger.info("Creating virtual environment")
        run_command('"{}" -m venv "{}" --system-site-packages'.format(base_venv, venv_dir))
        desktop_object.logger.info("Virtual environment created.")

    is_installed = False
    script_file = None
    if os.path.isdir(os.path.normpath(os.path.join(package_dir, toolkit_info["script"]))):
        script_file = os.path.normpath(os.path.join(package_dir, toolkit_info["script"]))
    else:
        for dirpath, dirnames, _ in os.walk(package_dir):
            if "site-packages" in dirnames:
                script_file = os.path.normpath(os.path.join(dirpath, "site-packages", toolkit_info["script"]))
                break
    if os.path.isfile(script_file):
        is_installed = True
    if wheel_toolkit:
        wheel_toolkit = os.path.normpath(wheel_toolkit)
    desktop_object.logger.info("Installing dependencies")
    if install and wheel_toolkit and os.path.exists(wheel_toolkit):
        desktop_object.logger.info("Starting offline installation")
        if is_installed:
            run_command('"{}" uninstall --yes {}'.format(pip_exe, toolkit_info["pip"]))
        import zipfile

        unzipped_path = os.path.join(
            os.path.dirname(wheel_toolkit), os.path.splitext(os.path.basename(wheel_toolkit))[0]
        )
        if os.path.exists(unzipped_path):
            shutil.rmtree(unzipped_path, ignore_errors=True)
        with zipfile.ZipFile(wheel_toolkit, "r") as zip_ref:
            zip_ref.extractall(unzipped_path)

        package_name = toolkit_info["package"]
        run_command(
            '"{}" install --no-cache-dir --no-index --find-links={} {}'.format(pip_exe, unzipped_path, package_name)
        )
    elif install and not is_installed:
        # Install the specified package
        run_command('"{}" --default-timeout=1000 install {}'.format(pip_exe, toolkit_info["pip"]))
    elif not install and is_installed:
        # Uninstall toolkit
        run_command('"{}" --default-timeout=1000 uninstall -y {}'.format(pip_exe, toolkit_info["package"]))
    elif install and is_installed:
        # Update toolkit
        run_command('"{}" --default-timeout=1000 install {} -U'.format(pip_exe, toolkit_info["pip"]))
    else:
        desktop_object.logger.info("Incorrect input")
        return
    toolkit_dir = os.path.join(desktop_object.personallib, "Toolkits")
    tool_dir = os.path.join(toolkit_dir, product_name, toolkit_info["name"])

    script_image = os.path.abspath(
        os.path.join(os.path.dirname(pyaedt.workflows.__file__), product_name.lower(), toolkit_info["icon"])
    )

    if install:
        if not os.path.exists(tool_dir):
            # Install toolkit inside AEDT
            add_script_to_menu(
                name=toolkit_info["name"],
                script_file=script_file,
                icon_file=script_image,
                product=product_name,
                template_file=toolkit_info.get("template", "run_pyaedt_toolkit_script"),
                copy_to_personal_lib=True,
                executable_interpreter=python_exe,
                personal_lib=desktop_object.personallib,
                aedt_version=desktop_object.aedt_version_id,
            )
            desktop_object.logger.info("{} installed".format(toolkit_info["name"]))
            if version > "232":
                desktop_object.odesktop.RefreshToolkitUI()
    else:
        if os.path.exists(tool_dir):
            # Install toolkit inside AEDT
            remove_script_from_menu(
                desktop_object=desktop_object,
                name=toolkit_info["name"],
                product=product_name,
            )
            desktop_object.logger.info("Installing dependencies")
            desktop_object.logger.info("{} uninstalled".format(toolkit_info["name"]))


def remove_script_from_menu(desktop_object, name, product="Project"):
    """Remove a toolkit script from the menu.

    Parameters
    ----------
    desktop_object : :class:pyaedt.desktop.Desktop
        Desktop object.
    name : str
        Name of the toolkit to remove.
    product : str, optional
        Product to which the toolkit applies. The default is ``"Project"``, in which case
        it applies to all designs. You can also specify a product, such as ``"HFSS"``.

    Returns
    -------
    bool
    """
    product = __tab_map(product)
    toolkit_dir = os.path.join(desktop_object.personallib, "Toolkits")
    aedt_version = desktop_object.aedt_version_id
    tool_dir = os.path.join(toolkit_dir, product, name)
    shutil.rmtree(tool_dir, ignore_errors=True)
    if aedt_version >= "2023.2":
        remove_xml_tab(os.path.join(toolkit_dir, product), name)
    desktop_object.logger.info("{} toolkit removed successfully.".format(name))
    return True


def __exe():
    if not is_linux:
        return ".exe"
    return ""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import sys
import warnings

if os.name == "nt":
    os.environ["PYTHONMALLOC"] = "malloc"

LATEST_DEPRECATED_PYTHON_VERSION = (3, 7)


def deprecation_warning():
    """Warning message informing users that some Python versions are deprecated in PyAEDT."""
    # Store warnings showwarning
    existing_showwarning = warnings.showwarning

    # Define and use custom showwarning
    def custom_show_warning(message, category, filename, lineno, file=None, line=None):
        """Custom warning used to remove <stdin>:loc: pattern."""
        print("{}: {}".format(category.__name__, message))

    warnings.showwarning = custom_show_warning

    current_version = sys.version_info[:2]
    if current_version <= LATEST_DEPRECATED_PYTHON_VERSION:
        str_current_version = "{}.{}".format(*sys.version_info[:2])
        warnings.warn(
            "Current python version ({}) is deprecated in PyAEDT. We encourage you "
            "to upgrade to the latest version to benefit from the latest features "
            "and security updates.".format(str_current_version),
            PendingDeprecationWarning,
        )

    # Restore warnings showwarning
    warnings.showwarning = existing_showwarning


deprecation_warning()

#

pyaedt_path = os.path.dirname(__file__)
__version__ = "0.9.7"
version = __version__

#
if not ("IronPython" in sys.version or ".NETFramework" in sys.version):  # pragma: no cover
    import pyaedt.downloads as downloads
from pyaedt.edb import Edb  # nosec
from pyaedt.edb import Siwave  # nosec
from pyaedt.generic import constants
import pyaedt.generic.DataHandlers as data_handler
from pyaedt.generic.design_types import Circuit
from pyaedt.generic.design_types import Desktop
from pyaedt.generic.design_types import Emit
from pyaedt.generic.design_types import Hfss
from pyaedt.generic.design_types import Hfss3dLayout
from pyaedt.generic.design_types import Icepak
from pyaedt.generic.design_types import Maxwell2d
from pyaedt.generic.design_types import Maxwell3d
from pyaedt.generic.design_types import MaxwellCircuit
from pyaedt.generic.design_types import Mechanical
from pyaedt.generic.design_types import Q2d
from pyaedt.generic.design_types import Q3d
from pyaedt.generic.design_types import Rmxprt
from pyaedt.generic.design_types import Simplorer
from pyaedt.generic.design_types import TwinBuilder
from pyaedt.generic.design_types import get_pyaedt_app
from pyaedt.generic.design_types import launch_desktop
import pyaedt.generic.general_methods as general_methods
from pyaedt.generic.general_methods import _retry_ntimes
from pyaedt.generic.general_methods import generate_unique_folder_name
from pyaedt.generic.general_methods import generate_unique_name
from pyaedt.generic.general_methods import generate_unique_project_name
from pyaedt.generic.general_methods import inside_desktop
from pyaedt.generic.general_methods import is_ironpython
from pyaedt.generic.general_methods import is_linux
from pyaedt.generic.general_methods import is_windows
from pyaedt.generic.general_methods import online_help
from pyaedt.generic.general_methods import pyaedt_function_handler
from pyaedt.generic.general_methods import settings
from pyaedt.misc import current_student_version
from pyaedt.misc import current_version
from pyaedt.misc import installed_versions



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import random
import re
import sys
import time
import warnings

from pyaedt.generic.constants import AEDT_UNITS
from pyaedt.generic.general_methods import is_linux
from pyaedt.generic.general_methods import pyaedt_function_handler
from pyaedt.generic.settings import settings
from pyaedt.modeler.cad.Modeler import Modeler

if (3, 8) < sys.version_info < (3, 12):
    from pyaedt.modeler.circuits.PrimitivesEmit import EmitComponent
    from pyaedt.modeler.circuits.PrimitivesEmit import EmitComponents
else:  # pragma: no cover
    warnings.warn("Emit API is only available for Python 3.8+,<3.12.")
from pyaedt.modeler.circuits.PrimitivesMaxwellCircuit import MaxwellCircuitComponents
from pyaedt.modeler.circuits.PrimitivesNexxim import NexximComponents
from pyaedt.modeler.circuits.PrimitivesTwinBuilder import TwinBuilderComponents
from pyaedt.modeler.circuits.object3dcircuit import CircuitComponent
from pyaedt.modeler.circuits.object3dcircuit import Wire
from pyaedt.modeler.pcb.Primitives3DLayout import Primitives3DLayout
from pyaedt.modules.LayerStackup import Layers


class ModelerCircuit(Modeler):
    """ModelerCircuit class.

    Parameters
    ----------
    app : :class:`pyaedt.application.AnalysisNexxim.FieldAnalysisCircuit`

    Examples
    --------
    >>> from pyaedt import Circuit
    >>> app = Circuit()
    >>> my_modeler = app.modeler
    """

    def __init__(self, app):
        app.logger.reset_timer()
        self._app = app
        self._schematic_units = "meter"
        Modeler.__init__(self, app)
        self.wire = Wire(self)
        app.logger.info_timer("ModelerCircuit class has been initialized!")

    @property
    def o_def_manager(self):
        """AEDT Definition manager."""
        return self._app.odefinition_manager

    @property
    def schematic_units(self):
        """Schematic units.

        Options are ``"mm"``, ``"mil"``, ``"cm"`` and all other metric and imperial units.
        The default is ``"meter"``.
        """
        return self._schematic_units

    @schematic_units.setter
    def schematic_units(self, value):
        if value in list(AEDT_UNITS["Length"].keys()):
            self._schematic_units = value
        else:
            self.logger.error("The unit %s is not supported.", value)

    @property
    def o_component_manager(self):
        """Component manager object."""
        return self._app.o_component_manager

    @property
    def o_model_manager(self):
        """Model manager object."""
        return self._app.o_model_manager

    @property
    def oeditor(self):
        """Oeditor Module.

        References
        ----------

        >>> oEditor = oDesign.SetActiveEditor("SchematicEditor")"""
        return self._app.oeditor

    @pyaedt_function_handler()
    def zoom_to_fit(self):
        """Zoom To Fit.

        References
        ----------

        >>> oEditor.ZoomToFit
        """
        self.oeditor.ZoomToFit()

    @pyaedt_function_handler(
        firstcomponent="starting_component",
        secondcomponent="ending_component",
        pinnum_first="pin_starting",
        pinnum_second="pin_ending",
    )
    def connect_schematic_components(self, starting_component, ending_component, pin_starting=2, pin_ending=1):
        """Connect schematic components.

        Parameters
        ----------
        starting_component : str
           Starting (right) component.
        ending_component : str
           Ending (left) component for the connection line.
        pin_starting : str, optional
             Number of the pin at which to terminate the connection from the right end of the
             starting component. The default is ``2``.
        pin_ending : str, optional
             Number of the pin at which to terminate the connection from the left end of the
             ending component. The default is ``1``.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.

        References
        ----------

        >>> oEditor.CreateWire
        """
        if self._app.design_type == "Maxwell Circuit":
            components = self.schematic.components
            obj1 = components[starting_component]
        else:
            components = self.components
            obj1 = components[starting_component]
        if "Port" in obj1.composed_name:
            pos1 = self.oeditor.GetPropertyValue("BaseElementTab", obj1.composed_name, "Component Location").split(", ")
            pos1 = [float(i.strip()[:-3]) * 0.0000254 for i in pos1]
            if "GPort" in obj1.composed_name:
                pos1[1] += 0.00254
        else:
            if self._app.design_type == "Maxwell Circuit":
                pos1 = [float(re.sub(r"[^0-9.\-]", "", x)) * 0.0000254 for x in obj1.location]
            else:
                pins1 = components.get_pins(starting_component)
                pos1 = components.get_pin_location(starting_component, pins1[pin_starting - 1])
        obj2 = components[ending_component]
        if "Port" in obj2.composed_name:
            pos2 = self.oeditor.GetPropertyValue("BaseElementTab", obj2.composed_name, "Component Location").split(", ")
            pos2 = [float(i.strip()[:-3]) * 0.0000254 for i in pos2]
            if "GPort" in obj2.composed_name:
                pos2[1] += 0.00254

        else:
            if self._app.design_type == "Maxwell Circuit":
                pos2 = [float(re.sub(r"[^0-9.\-]", "", x)) * 0.0000254 for x in obj2.location]
            else:
                pins2 = components.get_pins(ending_component)
                pos2 = components.get_pin_location(ending_component, pins2[pin_ending - 1])
        try:
            self.schematic.create_wire([pos1, pos2])
            return True
        except Exception:
            return False

    @pyaedt_function_handler()
    def create_text(
        self,
        text,
        x_origin=0,
        y_origin=0,
        text_size=12,
        text_angle=0,
        text_color=0,
        show_rect=False,
        x1=0,
        y1=0,
        x2=0,
        y2=0,
        rect_line_width=0,
        rect_border_color=0,
        rect_fill=0,
        rect_color=0,
    ):
        """Draw Text.

        Parameters
        ----------
        text : string
            Text to display.
        x_origin : float, optional
            x origin coordinate of the text box.
            Default value is ``0``.
        y_origin : float, optional
            y origin coordinate of the text box .
            Default value is ``0``.
        text_size : int, optional
            Size of text.
            Default value is ``12``.
        text_angle : float, optional
            Angle of text.
            Default value is ``0``.
        text_color : int, optional
            The RGB value of the text color.
            Default value is ``0``.
        show_rect : bool, optional
            Show rectangle.
            Default value is ``False``.
        x1 : float, optional
            The text rectangle left X value, in meters.
            Default value is ``0``.
        y1 : float, optional
            The text rectangle upper Y value, in meters.
            Default value is ``0``.
        x2 : float, optional
            The text rectangle right X value, in meters.
            Default value is ``0``.
        y2 : float, optional
            The text rectangle lower Y value, in meters.
            Default value is ``0``.
        rect_line_width : float, optional
            The width of the rectangle border, in meters.
            Default value is ``0``.
        rect_border_color : int, optional
            The RGB value of the rectangle border color.
            Default value is ``0``.
        rect_fill : int, optional
            The rectangle fill pattern id.
            Available values are: 0 = hollow, 1 = solid, 2 = NEDiagonal, 3 = OrthoCross,
            4 = DiagCross, 5 = NWDiagonal, 6 = Horizontal, 7 = Vertical.
            Default value is ``0``.
        rect_color : int, optional
            The RGB value of the rectangle fill color.
            Default value is ``0``.


        Returns
        -------
        str
             Unique id of the created object when successful, ``False`` when failed.

        References
        ----------

        >>> oEditor.CreateText

        """
        fill = {
            0: "Hollow",
            1: "Solid",
            2: "NE Diagonal",
            3: "Orthogonal Cross",
            4: "Diagonal Cross",
            5: "NW Diagonal",
            6: "Horizontal",
            7: "Vertical",
        }
        x_origin, y_origin = self.schematic._convert_point_to_meter([x_origin, y_origin])
        x1, y1 = self.schematic._convert_point_to_meter([x1, y1])
        x2, y2 = self.schematic._convert_point_to_meter([x2, y2])

        element_ids = []
        for el in self.oeditor.GetAllGraphics():
            element_ids.append(int(el.split("@")[1]))
        system_random = random.SystemRandom()
        text_id = system_random.randint(20000, 23000)
        while text_id in element_ids:
            text_id = system_random.randint(20000, 23000)
        args = [
            "NAME:TextData",
            "X:=",
            x_origin,
            "Y:=",
            y_origin,
            "Size:=",
            12,
            "Angle:=",
            0,
            "Text:=",
            text,
            "Color:=",
            0,
            "Id:=",
            text_id,
            "ShowRect:=",
            show_rect,
            "X1:=",
            x1,
            "Y1:=",
            y1,
            "X2:=",
            x2,
            "Y2:=",
            y2,
            "RectLineWidth:=",
            0,
            "RectBorderColor:=",
            0,
            "RectFill:=",
            0,
            "RectColor:=",
            0,
        ]
        a = ["NAME:Attributes", "Page:=", 1]
        try:
            text_out = self.oeditor.CreateText(args, a)
            if isinstance(text_color, (tuple, list)):
                r, g, b = text_color
            else:
                r = (text_color >> 16) & 0xFF
                g = (text_color >> 8) & 0xFF
                b = (text_color >> 0) & 0xFF
            if isinstance(rect_color, (tuple, list)):
                r2, g2, b2 = rect_color
            else:
                r2 = (rect_color >> 16) & 0xFF
                g2 = (rect_color >> 8) & 0xFF
                b2 = (rect_color >> 0) & 0xFF
            if isinstance(rect_border_color, (tuple, list)):
                r3, g3, b3 = rect_border_color
            else:
                r3 = (rect_border_color >> 16) & 0xFF
                g3 = (rect_border_color >> 8) & 0xFF
                b3 = (rect_border_color >> 0) & 0xFF
            self.change_text_property(str(text_id), "Color", [r, g, b])
            self.change_text_property(str(text_id), "Angle", self._arg_with_dim(text_angle, "deg"))
            self.change_text_property(str(text_id), "DisplayRectangle", show_rect)
            if show_rect:
                self.change_text_property(str(text_id), "Rectangle Color", [r2, g2, b2])
                self.change_text_property(
                    str(text_id), "Rectangle BorderWidth", self._arg_with_dim(rect_line_width, "mil")
                )
                self.change_text_property(str(text_id), "Rectangle BorderColor", [r3, g3, b3])
                self.change_text_property(str(text_id), "Rectangle FillStyle", fill[rect_fill])
            return text_out
        except Exception:
            return False

    @pyaedt_function_handler(property_id="assignment", property_name="name", property_value="value")
    def change_text_property(self, assignment, name, value):
        """Change an oeditor property.

        Parameters
        ----------
        assignment : str
            Object id.
        name : str
            Name of the property. For example, ``Text``.
        value : str, list, int
            Value of the property. It can be a string, an int for a single value, a list of three elements for
            ``[r,g,b]`` color values or a list of two elements for ``[x, y]`` coordinates.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.

        References
        ----------

        >>> oEditor.ChangeProperty
        """
        graphics_id = [id.split("@")[1] for id in self.oeditor.GetAllGraphics()]
        if assignment not in graphics_id:
            self.logger.error("Invalid id.")
            return False
        if isinstance(value, list) and len(value) == 3:
            if not isinstance(value[0], int) or not isinstance(value[1], int) or not isinstance(value[2], int):
                self.logger.error("Invalid RGB values for color")
                return False
            self.oeditor.ChangeProperty(
                [
                    "NAME:AllTabs",
                    [
                        "NAME:BaseElementTab",
                        ["NAME:PropServers", "SchObj@" + assignment],
                        [
                            "NAME:ChangedProps",
                            [
                                "NAME:" + name,
                                "R:=",
                                value[0],
                                "G:=",
                                value[1],
                                "B:=",
                                value[2],
                            ],
                        ],
                    ],
                ]
            )
        elif isinstance(value, list) and len(value) == 2:
            xpos = self._arg_with_dim(value[0])
            ypos = self._arg_with_dim(value[1])
            self.oeditor.ChangeProperty(
                [
                    "NAME:AllTabs",
                    [
                        "NAME:BaseElementTab",
                        ["NAME:PropServers", "SchObj@" + assignment],
                        ["NAME:ChangedProps", ["NAME:" + name, "X:=", xpos, "Y:=", ypos]],
                    ],
                ]
            )
        elif isinstance(value, bool):
            self.oeditor.ChangeProperty(
                [
                    "NAME:AllTabs",
                    [
                        "NAME:BaseElementTab",
                        ["NAME:PropServers", "SchObj@" + assignment],
                        ["NAME:ChangedProps", ["NAME:" + name, "Value:=", value]],
                    ],
                ]
            )
        elif isinstance(value, (str, float, int)):
            self.oeditor.ChangeProperty(
                [
                    "NAME:AllTabs",
                    [
                        "NAME:BaseElementTab",
                        ["NAME:PropServers", "SchObj@" + assignment],
                        ["NAME:ChangedProps", ["NAME:" + name, "Value:=", value]],
                    ],
                ]
            )
        else:
            self.logger.error("Wrong Property Value")
            return False
        self.logger.debug("Property {} Changed correctly.".format(name))
        return True

    @pyaedt_function_handler()
    def _get_components_selections(self, selections, return_as_list=True):
        sels = []
        if not isinstance(selections, list):
            selections = [selections]
        for sel in selections:
            if isinstance(sel, int):
                sels.append(self.schematic.components[sel].composed_name)
            elif isinstance(sel, CircuitComponent):
                sels.append(sel.composed_name)
            else:
                for el in list(self.schematic.components.values()):
                    if sel in [el.InstanceName, el.composed_name, el.name]:
                        sels.append(el.composed_name)
        if not return_as_list:
            return ", ".join(sels)
        return sels

    @pyaedt_function_handler()
    def _arg_with_dim(self, value, units=None):
        if units is None:
            units = self.schematic_units
        if isinstance(value, str):
            try:
                float(value)
                val = "{0}{1}".format(value, units)
            except Exception:
                val = value
        else:
            val = "{0}{1}".format(value, units)
        return val


class ModelerNexxim(ModelerCircuit):
    """ModelerNexxim class.

    Parameters
    ----------
    app : :class:`pyaedt.application.AnalysisNexxim.FieldAnalysisCircuit`

    """

    def __init__(self, app):
        self._app = app
        ModelerCircuit.__init__(self, app)
        self._schematic = NexximComponents(self)
        self._layouteditor = None
        self.layers = Layers(self, roughnessunits="um")
        self._primitives = Primitives3DLayout(app)
        self.logger.info("ModelerNexxim class has been initialized!")

    @property
    def layouteditor(self):
        """Return the Circuit Layout Editor.

        References
        ----------

        >>> oDesign.SetActiveEditor("Layout")
        """
        return self._app.layouteditor

    @property
    def schematic(self):
        """Schematic Component.

        Returns
        -------
        :class:`pyaedt.modeler.circuits.PrimitivesNexxim.NexximComponents`
        """
        return self._schematic

    @property
    def components(self):
        """Schematic Component.

        .. deprecated:: 0.4.13
           Use :func:`Circuit.modeler.schematic` instead.

        Returns
        -------
        :class:`pyaedt.modeler.circuits.PrimitivesNexxim.NexximComponents`
        """
        return self._schematic

    @property
    def edb(self):
        """EDB.

        Returns
        -------
        :class:`pyaedt.Edb`
            edb_core object if it exists.

        """
        # TODO Check why it crashes when multiple circuits are created
        return None

    @property
    def model_units(self):
        """Layout model units.

        References
        ----------

        >>> oEditor.GetActiveUnits
        >>> oEditor.SetActiveUnits
        """
        active_units = self.layouteditor.GetActiveUnits()
        if is_linux and settings.aedt_version == "2024.1":
            time.sleep(1)
            self._app.odesktop.CloseAllWindows()
        return active_units

    @property
    def layout(self):
        """Primitives.

        Returns
        -------
        :class:`pyaedt.modeler.Primitives3DLayout.Primitives3DLayout`

        """
        if self._app.design_type == "Twin Builder":
            return
        return self._primitives

    @property
    def primitives(self):
        """Primitives.

        .. deprecated:: 0.4.13
           Use :func:`Circuit.modeler.layout` instead.

        Returns
        -------
        :class:`pyaedt.modeler.Primitives3DLayout.Primitives3DLayout`

        """
        return self._primitives

    @model_units.setter
    def model_units(self, units):
        """Set the model units as a string e.g. "mm"."""
        assert units in AEDT_UNITS["Length"], "Invalid units string {0}".format(units)
        self.oeditor.SetActivelUnits(["NAME:Units Parameter", "Units:=", units, "Rescale:=", False])

    @pyaedt_function_handler(selections="assignment", pos="offset")
    def move(self, assignment, offset, units=None):
        """Move the selections by the specified ``[x, y]`` coordinates.

        Parameters
        ----------
        assignment : list
            List of the selections.
        offset : list
            Offset for the ``[x, y]`` axis.
        units : str
            Units of the movement. The default is ``meter``. If ``None``, ``schematic_units`` are used.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.

        References
        ----------

        >>> oEditor.Move
        """
        sels = self._get_components_selections(assignment)
        if not sels:
            self.logger.error("No Component Found.")
            return False
        if units:
            x_location = AEDT_UNITS["Length"][units] * float(offset[0])
            y_location = AEDT_UNITS["Length"][units] * float(offset[1])
        else:
            x_location = AEDT_UNITS["Length"][self.schematic_units] * float(offset[0])
            y_location = AEDT_UNITS["Length"][self.schematic_units] * float(offset[1])
        self.oeditor.Move(
            ["NAME:Selections", "Selections:=", sels],
            [
                "NAME:MoveParameters",
                "xdelta:=",
                x_location,
                "ydelta:=",
                y_location,
                "Disconnect:=",
                False,
                "Rubberband:=",
                False,
            ],
        )
        return True

    @pyaedt_function_handler(selections="assignment")
    def rotate(self, assignment, degrees=90):
        """Rotate the selections by degrees.

        Parameters
        ----------
        assignment : list
            List of the selections.
        degrees : optional
            Angle rotation in degrees. The default is ``90``.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.

        References
        ----------

        >>> oEditor.Rotate
        """
        sels = self._get_components_selections(assignment)
        if not sels:
            self.logger.error("No Component Found.")
            return False

        self.oeditor.Rotate(
            ["NAME:Selections", "Selections:=", sels],
            [
                "NAME:RotateParameters",
                "Degrees:=",
                degrees,
                "Disconnect:=",
                False,
                "Rubberband:=",
                False,
            ],
        )
        return True


class ModelerTwinBuilder(ModelerCircuit):
    """ModelerTwinBuilder class.

    Parameters
    ----------
    app : :class:`pyaedt.application.AnalysisTwinBuilder.AnalysisTwinBuilder`

    """

    def __init__(self, app):
        self._app = app
        ModelerCircuit.__init__(self, app)
        self._components = TwinBuilderComponents(self)
        self.logger.info("ModelerTwinBuilder class has been initialized!")

    @property
    def components(self):
        """
        .. deprecated:: 0.4.13
           Use :func:`TwinBuilder.modeler.schematic` instead.

        """
        return self._components

    @property
    def schematic(self):
        """Schematic Object.

        Returns
        -------
        :class:`pyaedt.modeler.PrimitivesTwinBuilder.TwinBuilderComponents`

        """
        return self._components


class ModelerEmit(ModelerCircuit):
    """ModelerEmit class.

    Parameters
    ----------
    app : :class:`pyaedt.application.AnalysisEmit`

    """

    def __init__(self, app):
        self._app = app
        ModelerCircuit.__init__(self, app)
        self.components = EmitComponents(app, self)
        self.logger.info("ModelerEmit class has been initialized!")

    @pyaedt_function_handler()
    def _get_components_selections(self, selections, return_as_list=True):  # pragma: no cover
        sels = []
        if not isinstance(selections, list):
            selections = [selections]
        for sel in selections:
            if isinstance(sel, int):
                sels.append(self.schematic.components[sel].composed_name)
            elif isinstance(sel, (CircuitComponent, EmitComponent)):
                sels.append(sel.composed_name)
            else:
                for el in list(self.schematic.components.values()):
                    if sel in [el.InstanceName, el.composed_name, el.name]:
                        sels.append(el.composed_name)
        if not return_as_list:
            return ", ".join(sels)
        return sels


class ModelerMaxwellCircuit(ModelerCircuit):
    """ModelerMaxwellCircuit class.

    Parameters
    ----------
    app : :class:`pyaedt.application.AnalysisMaxwellCircuit`

    """

    def __init__(self, app):
        self._app = app
        ModelerCircuit.__init__(self, app)
        self._components = MaxwellCircuitComponents(self)
        self.logger.info("ModelerMaxwellCircuit class has been initialized!")

    @property
    def schematic(self):
        """Schematic Object.

        Returns
        -------
        :class:`pyaedt.modeler.PrimitivesMaxwellCircuit.MaxwellCircuitComponents`

        """
        return self._components



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os

import pyaedt
import pyaedt.workflows.hfss3dlayout
from pyaedt.workflows.misc import get_aedt_version
from pyaedt.workflows.misc import get_arguments
from pyaedt.workflows.misc import get_port
from pyaedt.workflows.misc import get_process_id
from pyaedt.workflows.misc import is_student

port = get_port()
version = get_aedt_version()
aedt_process_id = get_process_id()
is_student = is_student()

# Extension batch arguments
extension_arguments = {"choice": "Export to HFSS"}
extension_description = "Export layout to 3D Modeler"

suffixes = {"Export to HFSS": "HFSS", "Export to Q3D": "Q3D", "Export to Maxwell 3D": "M3D", "Export to Icepak": "IPK"}


def frontend():  # pragma: no cover

    import tkinter
    from tkinter import ttk

    import PIL.Image
    import PIL.ImageTk

    master = tkinter.Tk()

    master.geometry("400x150")

    master.title("Export to 3D")

    # Load the logo for the main window
    icon_path = os.path.join(os.path.dirname(pyaedt.workflows.__file__), "images", "large", "logo.png")
    im = PIL.Image.open(icon_path)
    photo = PIL.ImageTk.PhotoImage(im)

    # Set the icon for the main window
    master.iconphoto(True, photo)

    # Configure style for ttk buttons
    style = ttk.Style()
    style.configure("Toolbutton.TButton", padding=6, font=("Helvetica", 10))

    var = tkinter.StringVar()
    label = tkinter.Label(master, textvariable=var, relief=tkinter.RAISED)
    var.set("Choose an option:")
    label.pack(pady=10)
    combo = ttk.Combobox(master, width=40)  # Set the width of the combobox
    combo["values"] = ("Export to HFSS", "Export to Q3D", "Export to Maxwell 3D", "Export to Icepak")
    combo.current(0)
    combo.pack(pady=10)

    combo.focus_set()

    def callback():
        master.choice_ui = combo.get()
        master.destroy()

    b = tkinter.Button(master, text="Export", width=40, command=callback)
    b.pack(pady=10)

    tkinter.mainloop()

    choice_ui = getattr(master, "choice_ui", extension_arguments["choice"])

    output_dict = {
        "choice": choice_ui,
    }
    return output_dict


def main(extension_args):
    choice = extension_args["choice"]

    app = pyaedt.Desktop(
        new_desktop=False,
        version=version,
        port=port,
        aedt_process_id=aedt_process_id,
        student_version=is_student,
    )

    active_project = app.active_project()
    active_design = app.active_design()

    project_name = active_project.GetName()

    if active_design.GetDesignType() in ["HFSS 3D Layout Design"]:
        design_name = active_design.GetName().split(";")[1]
    else:  # pragma: no cover
        app.logger.debug("Hfss 3D Layout project is needed.")
        app.release_desktop(False, False)
        raise Exception("Hfss 3D Layout project is needed.")

    h3d = pyaedt.Hfss3dLayout(project=project_name, design=design_name)
    setup = h3d.create_setup()
    suffix = suffixes[choice]

    if choice == "Export to Q3D":
        setup.export_to_q3d(h3d.project_file[:-5] + f"_{suffix}.aedt", keep_net_name=True)
    else:
        setup.export_to_hfss(h3d.project_file[:-5] + f"_{suffix}.aedt", keep_net_name=True)

    h3d.delete_setup(setup.name)

    h3d.save_project()

    if choice == "Export to Q3D":
        _ = pyaedt.Q3d(project=h3d.project_file[:-5] + f"_{suffix}.aedt")
    else:
        aedtapp = pyaedt.Hfss(project=h3d.project_file[:-5] + f"_{suffix}.aedt")
        aedtapp2 = None
        if choice == "Export to Maxwell 3D":
            aedtapp2 = pyaedt.Maxwell3d(project=aedtapp.project_name)
        elif choice == "Export to Icepak":
            aedtapp2 = pyaedt.Icepak(project=aedtapp.project_name)
        if aedtapp2:
            aedtapp2.copy_solid_bodies_from(aedtapp, no_vacuum=False, no_pec=False, include_sheets=True)
            aedtapp2.delete_design(aedtapp.design_name)
            aedtapp2.save_project()

    if not extension_args["is_test"]:  # pragma: no cover
        app.logger.info("Project generated correctly.")
        app.release_desktop(False, False)
    return True


if __name__ == "__main__":  # pragma: no cover
    args = get_arguments(extension_arguments, extension_description)

    # Open UI
    if not args["is_batch"]:  # pragma: no cover
        output = frontend()
        if output:
            for output_name, output_value in output.items():
                if output_name in extension_arguments:
                    args[output_name] = output_value

    main(args)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

spimsim_matlab_keywords_mapping = {
    # Matlab keyword -> SPIsim keyword
    # OP_IO_CTRL
    "BESSEL_THOMSON": "BESSEL_THOMSON",  # enable Bessel Thomsen filter for COM
    "Butterworth": "BUTTERWORTH",  # enable Butterworth filter for TDR, PTDR, and ERL
    "CDR": "CDR",  # CDR method, default is 'MM' (Mueller-Muller)
    "COM Pass threshold": "COM_PASS_THRESHOLD",  # the pass fail threshold for COM in dB
    "Enforce Causality DIFF_TOL": "EC_DIFF_TOL",
    # Difference Tolerance parameter for causality, Hard enforcement, 1e-4,Soft enforcement, 1e-3
    "Enforce Causality pulse start tolerance": "EC_PULSE_TOL",
    # Tolerance parameter for causality, Hard enforcement, 0.05, Soft enforcement, .01
    "Enforce Causality REL_TOL": "EC_REL_TOL",
    # Difference Tolerance parameter for causality, Hard enforcement, 1e-4,Soft enforcement, 1e-3
    "EH_MAX": "EH_MAX",  # used when PMD_type is C2M and is not really computed per spec
    "EH_MIN": "EH_MIN",  # used when PMD_type is C2M
    "ENFORCE_CAUSALITY": "ENFORCE_CAUSALITY",  # default is 0. Not recommended to use
    "ERL Pass threshold": "ERL_PASS_THRESHOLD",  # the pass fail threshold for ERL in dB
    "Error propagation COM margin": "ERR_PROPAGATION_COM_MARGIN",
    # Use to calculate  error propagation (not normally used)
    "Force PDF bin size": "FORCE_PDF_BIN_SIZE",  # do not use
    "Histogram_Window_Weight": "HISTOGRAM_WINDOW_WEIGHT",
    # Weighting for VEC and VEO are histogram processing. Type are Gaussian,Dual Rayleigh,Triangle, and Rectangle
    # (default)
    "Impulse response truncation threshold": "IMPRSP_TRUNC_THRESHOLD",  # Impulse response truncation threshold
    "Include PCB": "INCLUDE_PCB",  # Used to add a PCB one each side of the passed s-parameters.
    "nburst": "MAX_BURST_LEN",  # Use to calculate burst error rate (not normally used)
    "Min_VEO_Test": "MIN_VEO_TEST",
    # used when PMD_type is C2M. This allow EH to go blow EH_min. If set to Zero it is ignored
    "N_v": "N_V",  # number of UI used to compute Vf
    "Optimize_loop_speed_up": "OPTIMIZE_LOOP_SPEED_UP",
    # If set to 0 (or default) normal looping, If set to 1 loop speedup by slightly reducing PD Fbin and FIR_threshold
    # for optimize looping only
    "Port Order": "PORT_ORDER",  # s parameter port order [ tx+ tx- rx+ rx-]
    "RUNTAG": "RUNTAG",  # This string is appended to the beginning of results files
    "SBR_GEN_METHOD": "SBR_GEN_METHOD",  # Pulse generation method: use IBIS's slew rate or filtered rectangular pause
    "sigma_r": "SIGMA_R",  # sigma_r for 0.3ck Gaussian histogram window. Unit are UI. Preferred usage.
    "TDECQ": "TDECQ",  # Experimental, for only option is none (0) or vma. Default is 0.
    "TDR_W_TXPKG": "TDR_W_TXPKG",  # adds tx package for TDR, PTDR, and ERL. Default is 0.
    "USE_ETA0_PSD": "USE_ETA0_PSD",  # Used eta_0 PSD equation for sigma_n. Default is 0. Do not use.
    "VEC Pass threshold": "VEC_PASS_THRESHOLD",  # the pass fail threshold for VEC in dB only used when PMD_type is C2M
    # TABLE_93A1
    "A_fe": "A_FE",  # FEXT aggressor differential peak source output voltage (half of peak to peak)
    "A_ne": "A_NE",  # NEXT aggressor differential peak source output voltage (half of peak to peak)
    "A_v": "A_V",  # Victim differential peak source output voltage (half of peak to peak)
    "AC_CM_RMS": "AC_CM_RMS",
    # AC_CM_RMS is the CM BBN AWGN RMS at COM source point. Default is 0. Adds common mode noise source to the COM
    # signal path for the through channel
    "ACCM_MAX_Freq": "ACCM_MAX_FREQ",  # F max for integrating ACCM voltage in Hz. Default is fb
    "C_b": "C_B",  # C_b in nF (single sided)
    "C_d": "C_D",  # C_d in nF (single sided)
    "C_p": "C_P",  # C_p in nF (single sided)
    "C_V": "C_V",  # C_v in nF (via cap)  (single sided)
    "Delta_f": "DELTA_F",  # frequency step
    "DER_0": "DER_0",  # Target detector error ratio
    "f_b": "F_B",  # Baud (Signaling) rate in Gbaud
    "f_min": "F_MIN",  # minimum required frequency start for s parameters
    "f_v": "F_V",  # For FOM_ILD: Transition rate cutoff frequency for ICN/ILD calc in terms of fb
    "L": "L",  # number of symbols levels (PAM-4 is 4, NRZ is 2)
    "L_s": "L_S",  # L_s in nH (single sided)
    "Local Search": "LOCAL_SEARCH",  # Decreases COM compute time. Setting to 2 seems ok ,if 0 search is full grid
    "M": "M",  # Samples per UI
    "N_B_STEP": "N_B_STEP",  # Discretization of DFE, 0 disables and is not normally used
    "N_bx": "N_BX",  # Used for ERL to Compensate for a number of Ui associated with the DFE
    "R_0": "R_0",  # reference impedance
    "R_d": "R_D",  # Die source termination resistance  (single sided)
    "samples_for_C2M": "SAMPLES_FOR_C2M",  # Finer sampling in terms of samples per UI for c2m histogram analysis.
    "T_h": "T_O",  # superseded with T_O but is the internal values that is used. Do not use.
    "z_p (FEXT)": "Z_PFEXT",  # List of FEXT transmitter package trace lengths in mm, one per case
    "z_p (NEXT)": "Z_PNEXT",  # List of NEXT transmitter package trace lengths in mm, one per case
    "z_p (RX)": "Z_PRX",  # List of FEXT receiver package trace lengths in mm, one per case
    "z_p select": "Z_PSELECT",  # List of package length indexes used to run COM
    "z_p (TX)": "Z_PTX",  # List of victim transmitter package trace lengths in mm, one per case
    # TABLE_93A3
    "package_tl_gamma0_a1_a2": "PACKAGE_TL_GAMMA0_A1_A2",
    # Fitting parameters for package model per unit length. First element is in 1/mm and affects DC loss of package
    # model . Second element is in ns1/2/mm and affects loss proportional to sqrt(f). Third element is in ns/mm and
    # affects loss proportional to f.
    "package_tl_tau": "PACKAGE_TL_TAU",  # Package model transmission line delay ns/mm
    "package_Z_c": "PACKAGE_Z_C",  # Package model transmission line characteristic impedance [ Tx , Rx ]
    # TABLE_9212
    "board_tl_gamma0_a1_a2": "BOARD_TL_GAMMA0_A1_A2",
    # Fitting parameters for package model per unit length. First element is in 1/mm and affects DC loss of package
    # model . Second element is in ns1/2/mm and affects loss proportional to sqrt(f). Third element is in ns/mm and
    # affects loss proportional to f.
    "board_tl_tau": "BOARD_TL_TAU",  # Board model transmission line delay ns/mm
    "board_Z_c": "BOARD_Z_C",  # Board model transmission line characteristic impedance [ Tx , Rx ]
    "C_0": "CB0",  # If Include PCB is set to 1, near device single ended capacitance C0  in nF is added
    "C_1": "CB1",  # if Include PCB is set to 1, connector side single ended capacitance C1 in nF is added
    "z_bp (FEXT)": "Z_BPFEXT",  # Fext Assessor transmitter board trace lengths in mm
    "z_bp (NEXT)": "Z_BPNEXT",  # Next Assessor transmitter board trace lengths in mm
    "z_bp (RX)": "Z_BPRX",  # Victim receiver board trace lengths in mm
    "z_bp (TX)": "Z_BPTX",  # Victim transmitter board trace lengths in mm
    # TDR_ERL
    "AUTO_TFX": "AUTO_TFX",  # Mostly used for device ERL. If sent to 1 the fixture tfx will be estimated.
    "beta_x": "BETA_X",  # (for ERL) use default 0
    "BTorder": "BT_ORDER",  # Bessel function order
    "ERL": "ERL",  # Enables ERL. Needs TDR to be set as well.
    "ERL_ONLY": "ERL_ONLY",  # Compute ERL only
    "FIXTURE_BUILTIN_DELAY": "FIXTURE_BUILTIN_DELAY",  # built-in fixture delay
    "fixture delay time": "FIXTURE_DELAY_TIME",  # fixture delay time (for ERL)
    "N": "N",  # duration time in UI which is used for ERL (PTDR)
    "rho_x": "RHO_X",  # (for ERL) use default 0.618
    "TDR_DURATION": "TDR_DURATION",
    # only used if N*UI is longer than the TDR duration time.  Default is 5 times the raw s-parameter transit time.
    "TR_TDR": "TR_TDR",  # Gaussian shaped transition time for TDR source in ns
    "Turkey_Window": "TUKEY_WINDOW",  # required for ERL. Set to 1. Default is 0.
    "Z_T": "Z_T",  # single sided source termination reference resistance for TDR and ERL
    # ICN_PARAM
    "A_ft": "A_FT",  # FEXT aggressor amplitude for ICN. Defaults to A_fe if not specified
    "A_nt": "A_NT",  # NEXT aggressor amplitude for ICN. Defaults to A_ne if not specified
    "f_1": "F_1",  # start frequency for ICN and ILD calculations in GHz
    "f_2": "F_2",  # frequency in GHz for integration computation of ICN or FOM_Ild in GHz
    "f_f": "F_F",  # For ICN: Fext transition rate cut off frequency for ICN calc in terms of fb
    "f_n": "F_N",  # For ICN: Next transition rate cut off frequency for ICN calc in terms of fb
    # FILTER_EQ
    "b_max(1)": "B_MAX1",  # DFE magnitude limit, first coefficient(ignored if Nb=0)
    "b_max(2..N_b)": "B_MAX2_N_B",
    # DFE magnitude limit, second coefficient and on (ignored if Nb<2). Can be a regular expression
    "b_min(1)": "B_MIN1",  # DFE negative magnitude limit. If not specified it defaults to -bmax.
    "b_min(2..N_b)": "B_MIN2_N_B",  # DFE negative magnitude limit, if not specified it defaults to -b_max(2..N_b)
    "c(-1)": "C_1",  # TX equalizer pre cursor tap -1
    "c(-2)": "C_2",  # TX equalizer pre cursor tap -2
    "c(-3)": "C_3",  # TX equalizer pre cursor tap -3
    "c(-4)": "C_4",  # TX equalizer pre cursor tap -4
    "c(0)": "C0",
    # TX equalizer cursor minimum value (actual value is calculated as 1-sum(abs(tap)), Grid seat ignored for when C(0)
    # is below this value
    "c(1)": "C1",  # TX equalizer post cursor tap 1
    "c(2)": "C2",  # TX equalizer post cursor tap 2
    "c(3)": "C3",  # TX equalizer post cursor tap 3
    "CTLE_TYPE": "CTLE_TYPE",  # Sets the CTLE type default is poles and zeros (i.e. not a list of poles as in 120e)
    "F_HP_P": "F_HP_P",  # CFT pole fp2 is in GHz. Normally a list for 120e. Not normally use elsewhere.
    "f_HP_PZ": "F_HP_PZ",  # CFT pole zero pair in GHz for low frequency CTF
    "F_HP_Z": "F_HP_Z",  # CFT pole zero pair in GHz for low frequency CTF
    "f_p1": "F_P1",  # CTLE pole 1 in GHz
    "f_p2": "F_P2",  # CTLE pole 2 in GHz
    "f_r": "F_R",  # Receiver filter in COM and in ICN/FOM_ILD calcs in terms of fb
    "f_z": "F_Z",  # CTLE zero in GHz
    "g_DC": "G_DC",  # AC-DC gain list
    "g_DC_HP": "G_DC_HP",  # CTF AC-DC gain list (GDC2)
    "G_Qual": "G_QUAL",  # G_Qual are the dB ranges of g_DC g DC )which correspond tog_DC_HP (g DC2)
    "G2_Qual": "G2_QUAL",
    # G2_Qual limit values of g_DC_HP (g DC2 ) which corresponds to ranges of g_DC g DC specified with G_QUAL
    "GDC_MIN": "GDC_MIN",  # max ACDC gain, if 0 ignore
    "N_b": "N_B",  # Decision feedback fixed equalizer (DFE) length
    # NOISE_JITTER
    "A_DD": "A_DD",  # Normalized peak dual-Dirac noise, this is half of the total bound uncorrelated jitter (BUJ) in UI
    "eta_0": "ETA_0",
    # One-sided noise spectral density (V^2/GHz).Input referred noise at TP5. Input referred noise at TP5
    "R_LM": "R_LM",  # Ratio of level separation mismatch. Relevant when not PAM-2 (NRZ).
    "sigma_RJ": "SIGMA_RJ",  # rms of random jitter
    "SNR_TX": "SNR_TX",  # Transmitter SNDR noise in dB
    # RCV_NONSTD
    "FORCE_TR": "FORCE_TR",  # Included for earlier version support but should be set to 1 in most later config sheets.
    "Grr": "GRR",  # either do no use or set to 1 (for ERL)
    "GRR_LIMIT": "GRR_LIMIT",  # either do no use or set to 1 (for ERL)
    "Gx": "GX",  # ERL parameter param.Grr, This is used is the COM code
    "IDEAL_RX_TERM": "IDEAL_RX_TERM",
    # IDEAL_RX_TERM not supported, instead, set Zp,Cd, and Cp parameters to zero and Zp select to 1
    "IDEAL_TX_TERM": "IDEAL_TX_TERM",  # not supported, instead, set Zp,Cd, and Cp parameters to zero and Zp select to 1
    "INC_PACKAGE": "INC_PACKAGE",
    # warning: INC_PACKAGE=0 not fully supported, instead, set Zp,Cd, and Cp parameters to zero and Zp select to 1
    "INCLUDE_CTLE": "INCLUDE_CTLE",  # do not use
    "INCLUDE_TX_RX_FILTER": "INCLUDE_TX_RX_FILTER",  # do not use
    "kappa1": "KAPPA1",  # if set 0 reflection at tp0 are omitted from COM
    "kappa2": "KAPPA2",  # if set 0 reflection at tp5 are omitted from COM
    "RX_CALIBRATION": "RX_CALIBRATION",  # Turn on RX_Calibration loop
    "Sigma BBN step": "SIGMA_BBN_STEP",  # BBN step for Rx Calibration in volts. Defaults is 0.5e-3
    "T_r": "T_R",  # 20% to 80% transition time used for the Gaussian shaped source
    "T_R_FILTER_TYPE": "T_R_FILTER_TYPE",  # included for earlier version support. Not recommended to use.
    "T_R_MEAS_POINT": "T_R_MEAS_POINT",  # included for earlier version support. Not recommended to use.
    # FLOATING_TAPS
    "B_float_RSS_MAX": "B_FLOAT_RSS_MAX",  # floating DFE tap start for RSS floating tap limit
    "bmaxg": "BMAXG",  # max DFE value for floating taps
    "N_bf": "N_BF",  # number of taps in group
    "N_bg": "N_BG",  # number of group of floating tap. Used as a switch, 0 means no float
    "N_f": "N_F",  # UI span for floating taps. replaced by N_bmax
    "N_tail_start": "N_TAIL_START",  # start range for max RSS limit for DFE taps
    # RX_FFE
    "CURSOR_GAIN": "CURSOR_GAIN",  # only FFE and not supported
    "FFE_BACKOFF": "FFE_BACKOFF",
    # see if better zero forced solution is better by backing off the number specified FFE taps one at at time
    "FFE_MAIN_CUR_MIN": "FFE_MAIN_CUR_MIN",  # Rx FFE main cursor minimum
    "ffe_post_tap_len": "FFE_POST_TAP_LEN",  # Rx FFE post cursor tap length
    "FFE_POST_TAP1_MAX": "FFE_POST_TAP1_MAX",  # Rx FFE post cursor tap1 limit
    "ffe_pre_tap_len": "FFE_PRE_TAP_LEN",  # Rx FFE pre-cursor tap length
    "FFE_PRE_TAP1_MAX": "FFE_PRE_TAP1_MAX",  # Rx FFE pre cursor tap1 limit
    "FFE_TAP_STEP_SIZE": "FFE_TAP_STEP_SIZE",  # Rx FFE tap step size
    "FFE_TAPN_MAX": "FFE_TAPN_MAX",  # Rx FFE precursor tap N limit
    "Tukey_Window": "TUKEY_WINDOW",
    "TDR": "TDR",
    "NEXTARY": "NEXTARY",
    "HISTOGRAM_WINDOW_WEIGHT": "HISTOGRAM_WINDOW_WEIGHT",
    "OPTIMIZE_LOOP_SPEED_UP": "OPTIMIZE_LOOP_SPEED_UP",
    "EC_PULSE_TOL": "EC_PULSE_TOL",
    "VEC_PASS_THRESHOLD": "VEC_PASS_THRESHOLD",
    "KAPPA1": "KAPPA1",
    "VERSION": "VERSION",
    "PDF_BIN_SIZE": "PDF_BIN_SIZE",
    "RESULT_DIR": "RESULT_DIR",
    "G_QUAL": "G_QUAL",
    "T_O": "T_O",
    "TDR_BUTTERWORTH": "TDR_BUTTERWORTH",
    "GRR": "GRR",
    "SPECTAG": "SPECTAG",
    "FFE_PRE_TAP_LEN": "FFE_PRE_TAP_LEN",
    "FEXTARY": "FEXTARY",
    "GENHTML": "GENHTML",
    "SIGMA_R": "SIGMA_R",
    "FFE_POST_TAP_LEN": "FFE_POST_TAP_LEN",
    "MIN_VEO_TEST": "MIN_VEO_TEST",
    "ACCM_MAX_FREQ": "ACCM_MAX_FREQ",
    "N_V": "N_V",
    "BUTTERWORTH": "BUTTERWORTH",
    "IMPRSP_TRUNC_THRESHOLD": "IMPRSP_TRUNC_THRESHOLD",
    "EC_DIFF_TOL": "EC_DIFF_TOL",
    "C_4": "C_4",
    "ERR_PROPAGATION_COM_MARGIN": "ERR_PROPAGATION_COM_MARGIN",
    "EC_REL_TOL": "EC_REL_TOL",
    "MAX_BURST_LEN": "MAX_BURST_LEN",
    "FSTTHRU": "FSTTHRU",
    "BT_ORDER": "BT_ORDER",
    "PMD_TYPE": "PMD_TYPE",
    "C3": "C3",
    "C2": "C2",
    "F_1": "F_1",
    "G2_QUAL": "G2_QUAL",
    "NUMPORT": "NUMPORT",
    "NOISE_CREST_FACTOR": "NOISE_CREST_FACTOR",
    "THRUSNP": "THRUSNP",
    "SAMPLES_FOR_C2M": "SAMPLES_FOR_C2M",
    "GX": "GX",
}



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
This module contains these classes: `Mesh` and `Mesh3DOperation`.

This module provides all functionalities for creating and editing the mesh in the 3D tools.

"""

from __future__ import absolute_import  # noreorder

from collections import OrderedDict

from pyaedt.generic.DataHandlers import _dict2arg
from pyaedt.generic.general_methods import PropsManager
from pyaedt.generic.general_methods import generate_unique_name
from pyaedt.generic.general_methods import pyaedt_function_handler
from pyaedt.modules.Mesh import MeshProps


class Mesh3DOperation(PropsManager, object):
    """Mesh3DOperation class.

    Parameters
    ----------
    app :

    hfss_setup_name : str
        Name of the HFSS setup.
    name :

    props : dict
        Dictionary of the properties.

    """

    def __init__(self, app, hfss_setup_name, name, props):
        self.auto_update = True
        self._mesh3dlayout = app
        self.name = name
        self.props = MeshProps(self, props)
        self.hfss_setup_name = hfss_setup_name

    @pyaedt_function_handler()
    def _get_args(self, props=None):
        """Retrieve arguments.

        Parameters
        ----------
        props : dict
            Dictionary of the properties. The default is ``None``, in which
            case the default properties are retrieved.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.

        """
        if not props:
            props = self.props
        arg = ["NAME:" + self.name]
        _dict2arg(props, arg)
        return arg

    @pyaedt_function_handler()
    def create(self):
        """Create a mesh.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.

        References
        ----------

        >>> oModule.AddMeshOperation
        """
        self._mesh3dlayout.omeshmodule.AddMeshOperation(self.hfss_setup_name, self._get_args())
        return True

    @pyaedt_function_handler()
    def update(self, *args, **kwargs):
        """Update the mesh.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.

        References
        ----------

        >>> oModule.EditMeshOperation
        """
        self._mesh3dlayout.omeshmodule.EditMeshOperation(self.hfss_setup_name, self.name, self._get_args())
        return True

    @pyaedt_function_handler()
    def delete(self):
        """Delete the mesh.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.

        References
        ----------

        >>> oModule.DeleteMeshOperation
        """
        self._mesh3dlayout.omeshmodule.DeleteMeshOperation(
            self.hfss_setup_name,
            self.name,
        )

        return True


class Mesh3d(object):
    """Manages mesh operations for HFSS 3D Layout.

    Provides the main AEDT mesh functionality. The inherited class
    ``AEDTConfig`` contains all ``_desktop`` hierarchical calls needed by this class.

    Parameters
    ----------
    app : :class:`pyaedt.application.Analysis3DLayout.FieldAnalysis3DLayout`

    """

    def __init__(self, app):
        app.logger.reset_timer()
        self._app = app

        self.logger = self._app.logger
        self._odesign = self._app._odesign
        self.modeler = self._app.modeler
        self.id = 0

        self.meshoperations = self._get_design_mesh_operations()

        app.logger.info_timer("Mesh class has been initialized!")

    @pyaedt_function_handler()
    def generate_mesh(self, name):
        """Generate the mesh for a design.

        Parameters
        ----------
        name : str
            Name of the design.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.

        References
        ----------

        >>> oDesign.GenerateMesh
        """
        self._app.oanalysis.GenerateMesh([name])
        return True

    @property
    def omeshmodule(self):
        """AEDT Mesh Module.

        References
        ----------

        >>> oDesign.GetModule("SolveSetups")
        """
        return self._app.omeshmodule

    @pyaedt_function_handler(setup_name="setup", mesh_name="name")
    def delete_mesh_operations(self, setup, name):
        """Remove mesh operations from a setup.

        Parameters
        ----------
        setup : str
            Name of the setup.
        name : str
            Name of the mesh.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.

        References
        ----------

        >>> oModule.DeleteMeshOperation
        """
        for el in self.meshoperations:
            if el.hfss_setup_name == setup and el.name == name:
                el.delete()
                self.meshoperations.remove(el)

        return True

    @pyaedt_function_handler()
    def _get_design_mesh_operations(self):
        """Retrieve design mesh operations.

        Returns
        -------
        type
            Mesh operations object.

        """
        meshops = []
        try:
            for ds in self._app.design_properties["Setup"]["Data"]:
                if "MeshOps" in self._app.design_properties["Setup"]["Data"][ds]:
                    for ops in self._app.design_properties["Setup"]["Data"][ds]["MeshOps"]:
                        props = self._app.design_properties["Setup"]["Data"][ds]["MeshOps"][ops]
                        meshops.append(Mesh3DOperation(self, ds, ops, props))
        except Exception:
            pass
        return meshops

    @pyaedt_function_handler(
        setupname="setup",
        layer_name="layer",
        net_name="net",
        isinside="is_inside",
        maxlength="maximum_length",
        maxel="maximum_elements",
        meshop_name="name",
    )
    def assign_length_mesh(self, setup, layer, net, is_inside=True, maximum_length=1, maximum_elements=1000, name=None):
        """Assign mesh length.

        Parameters
        ----------
        setup : str
            Name of the HFSS setup to apply.
        layer : str
           Name of the layer.
        net : str
           Name of the net.
        is_inside : bool, optional
            Whether the mesh length is inside the selection. The default is ``True``.
        maximum_length : float, optional
            Maximum length of the element. The default is ``1`` When ``None``, this
            parameter is disabled.
        maximum_elements : int, optional
            Maximum number of elements. The default is ``1000``. When ``None``, this
            parameter is disabled.
        name : str, optional
            Name of the mesh operation. The default is ``None``.

        Returns
        -------
        type
            Mesh operation object.

        References
        ----------

        >>> oModule.AddMeshOperation
        """
        if name:
            for el in self.meshoperations:
                if el.name == name:
                    name = generate_unique_name(name)
        else:
            name = generate_unique_name("Length")

        if maximum_length is None:
            restrictlength = False
        else:
            restrictlength = True
        length = self.modeler.modeler_variable(maximum_length)

        if maximum_elements is None:
            restrictel = False
            numel = "1000"
        else:
            restrictel = True
            numel = str(maximum_elements)
        if maximum_length is None and maximum_elements is None:
            self.logger.error("mesh not assigned due to incorrect settings")
            return
        if isinstance(layer, list) and isinstance(net, list):
            assignment = OrderedDict({"MeshEntityInfo": []})
            for l, n in zip(layer, net):
                meshbody = OrderedDict({"Id": -1, "Nam": "", "Layer": l, "Net": n, "OrigNet": n})
                assignment["MeshEntityInfo"].append(
                    OrderedDict({"IsFcSel": False, "EntID": -1, "FcIDs": [], "MeshBody": meshbody, "BBox": []})
                )
        else:
            meshbody = OrderedDict({"Id": -1, "Nam": "", "Layer": layer, "Net": net, "OrigNet": net})
            assignment = OrderedDict(
                {
                    "MeshEntityInfo": OrderedDict(
                        {"IsFcSel": False, "EntID": -1, "FcIDs": [], "MeshBody": meshbody, "BBox": []}
                    )
                }
            )
        props = OrderedDict(
            {
                "Type": "LengthBased",
                "RefineInside": is_inside,
                "Enabled": True,
                "Assignment": assignment,
                "Region": "",
                "RestrictElem": restrictel,
                "NumMaxElem": numel,
                "RestrictLength": restrictlength,
                "MaxLength": length,
            }
        )

        mop = Mesh3DOperation(self, setup, name, props)
        mop.create()
        self.meshoperations.append(mop)
        return mop

    @pyaedt_function_handler(
        setupname="setup",
        layer_name="layer",
        net_name="net",
        skindepth="skin_depth",
        maxelements="maximum_elements",
        numlayers="layers_number",
        meshop_name="name",
    )
    def assign_skin_depth(
        self,
        setup,
        layer,
        net,
        skin_depth=1,
        maximum_elements=None,
        triangulation_max_length=0.1,
        layers_number="2",
        name=None,
    ):
        """Assign skin depth to the mesh.

        Parameters
        ----------
        setup : str
            Name of the setup.
        layer : str
            Name of the layer.
        net : str
            Name of the net.
        skin_depth : int, optional
            Depth of the skin. The default is ``1``.
        maximum_elements : float, optional
            Maximum element length. The default is ``None``, which disables this parameter.
        triangulation_max_length : float, optional
            Maximum surface triangulation length. The default is ``0.1``.
        layers_number : str, optional
            Number of layers. The default is ``"2"``.
        name : str, optional
             Name of the mesh operation. The default is ``None``.

        Returns
        -------
        type
            Mesh operation object.

        References
        ----------

        >>> oModule.AddMeshOperation
        """
        if name:
            for el in self.meshoperations:
                if el.name == name:
                    name = generate_unique_name(name)
        else:
            name = generate_unique_name("SkinDepth")

        if maximum_elements is None:
            restrictlength = False
            maximum_elements = "1000"
        else:
            restrictlength = True
        skin_depth = self.modeler.modeler_variable(skin_depth)
        triangulation_max_length = self.modeler.modeler_variable(triangulation_max_length)
        meshbody = OrderedDict({"Id": -1, "Nam": "", "Layer": layer, "Net": net, "OrigNet": net})
        assignment = OrderedDict(
            {
                "MeshEntityInfo": OrderedDict(
                    {"IsFcSel": False, "EntID": -1, "FcIDs": [], "MeshBody": meshbody, "BBox": []}
                )
            }
        )
        props = OrderedDict(
            {
                "Type": "SkinDepthLengthBased",
                "Enabled": True,
                "Assignment": assignment,
                "Region": "",
                "SkinDepth": skin_depth,
                "SurfTriMaxLength": triangulation_max_length,
                "NumLayers": layers_number,
                "RestrictElem": restrictlength,
                "NumMaxElem": maximum_elements,
            }
        )

        mop = Mesh3DOperation(self, setup, name, props)
        mop.create()
        self.meshoperations.append(mop)
        return mop



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
* * * This script is meant to run in IronPython within AEDT. * * *

It contains common methods for the PyAEDT panels.
"""
import os
import random
import string
import sys

from System.Windows.Forms import MessageBox
from System.Windows.Forms import MessageBoxButtons
from System.Windows.Forms import MessageBoxIcon

is_linux = os.name == "posix"


def sanitize_interpreter_path(interpreter_path, version):
    python_version = "3_10" if version > "231" else "3_7"
    if version > "231" and python_version not in interpreter_path:
        interpreter_path = interpreter_path.replace("3_7", "3_10")
    elif version <= "231" and python_version not in interpreter_path:
        interpreter_path = interpreter_path.replace("3_10", "3_7")
    return interpreter_path


def check_file(file_path, oDesktop):
    if not os.path.isfile(file_path):
        show_error(
            '"{}" does not exist. Install PyAEDT using the Python script installer from the PyAEDT '
            "documentation.".format(file_path),
            oDesktop,
        )
        return False
    return True


def get_linux_terminal():
    for terminal in ["x-terminal-emulator", "konsole", "xterm", "gnome-terminal", "lxterminal", "mlterm"]:
        term = which(terminal)
        if term:
            return term
    return None


def which(program):
    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, _ = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

    return None


def show_error(msg, oDesktop):
    oDesktop.AddMessage("", "", 2, str(msg))
    MessageBox.Show(str(msg), "Error", MessageBoxButtons.OK, MessageBoxIcon.Error)
    sys.exit()


def environment_variables(oDesktop):
    os.environ["PYAEDT_SCRIPT_PROCESS_ID"] = str(oDesktop.GetProcessID())
    version = str(oDesktop.GetVersion()[:6])
    os.environ["PYAEDT_SCRIPT_VERSION"] = version
    if version > "2023.1":
        os.environ["PYAEDT_SCRIPT_PORT"] = str(oDesktop.GetGrpcServerPort())
    else:
        os.environ["PYAEDT_SCRIPT_PORT"] = str(0)
    if "Ansys Student" in str(oDesktop.GetExeDir()):
        os.environ["PYAEDT_STUDENT_VERSION"] = "True"
    else:
        os.environ["PYAEDT_STUDENT_VERSION"] = "False"
    if is_linux:
        edt_root = os.path.normpath(oDesktop.GetExeDir())
        os.environ["ANSYSEM_ROOT{}".format(version)] = edt_root
        ld_library_path_dirs_to_add = [
            "{}/commonfiles/CPython/3_7/linx64/Release/python/lib".format(edt_root),
            "{}/commonfiles/CPython/3_10/linx64/Release/python/lib".format(edt_root),
            "{}/common/mono/Linux64/lib64".format(edt_root),
            "{}/Delcross".format(edt_root),
            "{}".format(edt_root),
        ]
        os.environ["LD_LIBRARY_PATH"] = ":".join(ld_library_path_dirs_to_add) + ":" + os.getenv("LD_LIBRARY_PATH", "")
        if version > "2023.1":
            os.environ["TCL_LIBRARY"] = os.path.join(
                "{}/commonfiles/CPython/3_10/linx64/Release/python/lib".format(edt_root), "tcl8.5"
            )
            os.environ["TK_LIBRARY"] = os.path.join(
                "{}/commonfiles/CPython/3_10/linx64/Release/python/lib".format(edt_root), "tk8.5"
            )
            os.environ["TKPATH"] = os.path.join(
                "{}/commonfiles/CPython/3_10/linx64/Release/python/lib".format(edt_root), "tk8.5"
            )
        else:
            os.environ["TCL_LIBRARY"] = os.path.join(
                "{}/commonfiles/CPython/3_7/linx64/Release/python/lib".format(edt_root), "tcl8.5"
            )
            os.environ["TK_LIBRARY"] = os.path.join(
                "{}/commonfiles/CPython/3_7/linx64/Release/python/lib".format(edt_root), "tk8.5"
            )
            os.environ["TKPATH"] = os.path.join(
                "{}/commonfiles/CPython/3_7/linx64/Release/python/lib".format(edt_root), "tk8.5"
            )


def generate_unique_name(root_name, suffix="", n=6):
    char_set = string.ascii_uppercase + string.digits
    unique_name = root_name + "_" + "".join(random.choice(char_set) for _ in range(n))  # nosec
    if suffix:
        unique_name += suffix
    return unique_name



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import re
import sys
import time

from pyaedt.circuit import Circuit
from pyaedt.desktop import Desktop

Emit = None
if not ("IronPython" in sys.version or ".NETFramework" in sys.version):  # pragma: no cover
    from pyaedt.emit import Emit
from pyaedt.generic.general_methods import is_linux
from pyaedt.generic.settings import settings
from pyaedt.hfss3dlayout import Hfss3dLayout
from pyaedt.hfss import Hfss
from pyaedt.icepak import Icepak
from pyaedt.maxwell import Maxwell2d
from pyaedt.maxwell import Maxwell3d
from pyaedt.maxwellcircuit import MaxwellCircuit
from pyaedt.mechanical import Mechanical
from pyaedt.q3d import Q2d
from pyaedt.q3d import Q3d
from pyaedt.rmxprt import Rmxprt
from pyaedt.twinbuilder import TwinBuilder

Simplorer = TwinBuilder


def launch_desktop(
    version=None,
    non_graphical=False,
    new_desktop=True,
    close_on_exit=True,
    student_version=False,
    machine="",
    port=0,
    aedt_process_id=None,
):
    """Initializes AEDT based on the inputs provided.

    Parameters
    ----------
    version : str, int, float, optional
        Version of AEDT to use. The default is ``None``, in which case the
        active setup or latest installed version is used.
    non_graphical : bool, optional
        Whether to launch AEDT in non-graphical mode. The default
        is ``False``, in which case AEDT is launched in graphical mode.
        This parameter is ignored when a script is launched within AEDT.
    new_desktop : bool, optional
        Whether to launch an instance of AEDT in a new thread, even if
        another instance of the ``specified_version`` is active on the machine.
        The default is ``False``.
    close_on_exit : bool, optional
        Whether to close AEDT on exit. The default is ``True``.
    student_version : bool, optional
        Whether to open the AEDT student version. The default is
        ``False``.
    machine : str, optional
        Machine name to connect the oDesktop session to. This parameters works only in 2022 R2
        and later. The remote server must be up and running with the command
        `"ansysedt.exe -grpcsrv portnum"`. If the machine is `"localhost"`, the server also
        starts if not present.
    port : int, optional
        Port number on which to start the oDesktop communication on the already existing server.
        This parameter is ignored when creating a new server. It works only in 2022 R2 and
        later. The remote server must be up and running with the command `"ansysedt.exe -grpcsrv portnum"`.
    aedt_process_id : int, optional
        Process ID for the instance of AEDT to point PyAEDT at. The default is
        ``None``. This parameter is only used when ``new_desktop = False``.

    Returns
    -------
    :class:`pyaedt.desktop.Desktop`


    Examples
    --------
    Launch AEDT 2021 R1 in non-graphical mode and initialize HFSS.

    >>> import pyaedt
    >>> desktop = pyaedt.launch_desktop("2022.2", non_graphical=True)
    PyAEDT INFO: pyaedt v...
    PyAEDT INFO: Python version ...
    >>> hfss = pyaedt.Hfss(design="HFSSDesign1")
    PyAEDT INFO: Project...
    PyAEDT INFO: Added design 'HFSSDesign1' of type HFSS.

    Launch AEDT 2021 R1 in graphical mode and initialize HFSS.

    >>> desktop = Desktop("2021.2")
    PyAEDT INFO: pyaedt v...
    PyAEDT INFO: Python version ...
    >>> hfss = pyaedt.Hfss(design="HFSSDesign1")
    PyAEDT INFO: No project is defined. Project...
    """
    d = Desktop(
        version=version,
        non_graphical=non_graphical,
        new_desktop=new_desktop,
        close_on_exit=close_on_exit,
        student_version=student_version,
        machine=machine,
        port=port,
        aedt_process_id=aedt_process_id,
    )
    return d


app_map = {
    "Maxwell 2D": Maxwell2d,
    "Maxwell 3D": Maxwell3d,
    "Maxwell Circuit": MaxwellCircuit,
    "Twin Builder": TwinBuilder,
    "Circuit Design": Circuit,
    "2D Extractor": Q2d,
    "Q3D Extractor": Q3d,
    "HFSS": Hfss,
    "Mechanical": Mechanical,
    "Icepak": Icepak,
    "Rmxprt": Rmxprt,
    "HFSS 3D Layout Design": Hfss3dLayout,
    "EMIT": Emit,
}


def get_pyaedt_app(project_name=None, design_name=None, desktop=None):
    """Gets the PyAEDT object with a given project name and design name.

    Parameters
    ----------
    project_name : str, optional
        Project name.
    design_name : str, optional
        Design name.
    desktop : :class:`pyaedt.desktop.Desktop`, optional
        Desktop class. The default is ``None``.

    Returns
    -------
    :def :`pyaedt.Hfss`
        Any of the Pyaedt App initialized.
    """
    from pyaedt.generic.desktop_sessions import _desktop_sessions

    odesktop = None
    process_id = None
    if desktop:
        odesktop = desktop.odesktop
        process_id = desktop.aedt_process_id
    elif _desktop_sessions and project_name:
        for desktop in list(_desktop_sessions.values()):
            if project_name in list(desktop.project_list()):
                odesktop = desktop.odesktop
                break
    elif _desktop_sessions:
        odesktop = list(_desktop_sessions.values())[-1]
    elif "oDesktop" in dir(sys.modules["__main__"]):  # ironpython
        odesktop = sys.modules["__main__"].oDesktop  # ironpython
    else:
        raise AttributeError("No Desktop Present.")
    if not process_id:
        process_id = odesktop.GetProcessID()
    if project_name and project_name not in odesktop.GetProjectList():
        raise AttributeError("Project  {} doesn't exist in current desktop.".format(project_name))
    if not project_name:
        oProject = odesktop.GetActiveProject()
    else:
        oProject = odesktop.SetActiveProject(project_name)
    if is_linux and settings.aedt_version == "2024.1":
        time.sleep(1)
        odesktop.CloseAllWindows()
    if not oProject:
        raise AttributeError("No project is present.")
    design_names = []
    deslist = list(oProject.GetTopDesignList())
    for el in deslist:
        m = re.search(r"[^;]+$", el)
        design_names.append(m.group(0))
    if design_name and design_name not in design_names:
        raise AttributeError("Design  {} doesn't exist in current project.".format(design_name))
    if not design_name:
        oDesign = oProject.GetActiveDesign()
    else:
        oDesign = oProject.SetActiveDesign(design_name)
    if is_linux and settings.aedt_version == "2024.1":
        time.sleep(1)
        odesktop.CloseAllWindows()
    if not oDesign:
        raise AttributeError("No design is present.")
    design_type = oDesign.GetDesignType()
    if design_type in list(app_map.keys()):
        version = odesktop.GetVersion().split(".")
        v = ".".join([version[0], version[1]])
        return app_map[design_type](project_name, design_name, version=v, aedt_process_id=process_id)
    return None



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from pyaedt.application.Analysis import Analysis
from pyaedt.generic.settings import settings


class AnalysisMaxwellCircuit(Analysis):
    """Provides the Maxwell Circuit (MaxwellCircuit) interface.
    Maxwell Circuit Editor has no setup, solution, analysis or postprocessor
    It is automatically initialized by Application call (Maxwell Circuit).
    Refer to Application function for inputs definition

    Parameters
    ----------
    projectname : str, optional
        Name of the project to select or the full path to the project
        or AEDTZ archive to open.  The default is ``None``, in which
        case an attempt is made to get an active project. If no
        projects are present, an empty project is created.
    designname : str, optional
        Name of the design to select. The default is ``None``, in
        which case an attempt is made to get an active design. If no
        designs are present, an empty design is created.
    version : str, int, float, optional
        Version of AEDT to use. The default is ``None``. If ``None``,
        the active setup is used or the latest installed version is
        used.
    NG : bool, optional
        Whether to launch AEDT in the non-graphical mode. The default
        is ``False``, in which case AEDT is launched in the graphical mode.
    new_desktop : bool, optional
        Whether to launch an instance of AEDT in a new thread, even if
        another instance of the ``specified_version`` is active on the
        machine.  The default is ``False``.
    close_on_exit : bool, optional
        Whether to release AEDT on exit. The default is ``False``.
    student_version : bool, optional
        Whether to open the AEDT student version. The default is ``False``.
    aedt_process_id : int, optional
        Only used when ``new_desktop = False``, specifies by process ID which instance
        of Electronics Desktop to point PyAEDT at.
    remove_lock : bool, optional
        Whether to remove lock to project before opening it or not.
        The default is ``False``, which means to not unlock
        the existing project if needed and raise an exception.
    """

    def __init__(
        self,
        application,
        projectname,
        designname,
        version=None,
        non_graphical=False,
        new_desktop=False,
        close_on_exit=False,
        student_version=False,
        machine="",
        port=0,
        aedt_process_id=None,
        remove_lock=False,
    ):
        Analysis.__init__(
            self,
            application,
            projectname,
            designname,
            None,
            None,
            version,
            non_graphical,
            new_desktop,
            close_on_exit,
            student_version,
            machine,
            port,
            aedt_process_id,
            remove_lock=remove_lock,
        )
        self._modeler = None
        if not settings.lazy_load:
            self._modeler = self.modeler

    @property
    def modeler(self):
        """Design oModeler.

        Returns
        -------
        :class:`pyaedt.modeler.schematic.ModelerMaxwellCircuit`
        """
        if self._modeler is None:
            from pyaedt.modeler.schematic import ModelerMaxwellCircuit

            self._modeler = ModelerMaxwellCircuit(self)
        return self._modeler



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from ctypes import CFUNCTYPE
from ctypes import PyDLL
from ctypes import c_bool
from ctypes import c_int
from ctypes import c_wchar_p
from ctypes import py_object
import os
import re
import socket
import time
import types

from pyaedt.generic.general_methods import GrpcApiError
from pyaedt.generic.general_methods import _retry_ntimes
from pyaedt.generic.general_methods import inclusion_list
from pyaedt.generic.general_methods import settings

logger = settings.logger


class AedtBlockObj(list):
    def GetName(self):
        if len(self) > 0:
            f = self[0]
            if isinstance(f, str):
                d = f.split(":")
                if len(d) == 2:
                    if d[0] == "NAME":
                        return d[1]

    def __GetValueIdxByKey__(self, keyName):
        for i in range(0, len(self) - 1):
            if isinstance(self[i], str):
                toks = self[i].split(":")
                if len(toks) == 2:
                    if toks[1] == "=" and toks[0] == keyName:
                        return i + 1
        raise GrpcApiError(keyName + " is not a key!")

    def __getitem__(self, idxOrKey):
        if isinstance(idxOrKey, str):
            idx = self.__GetValueIdxByKey__(idxOrKey)
            if idx != None:
                return super().__getitem__(idx)
        return super().__getitem__(idxOrKey)

    def __setitem__(self, idxOrKey, newVal):
        if isinstance(idxOrKey, int):
            if idxOrKey >= 0 or idxOrKey < len(self):
                oldItem = self.__getitem__(idxOrKey)
                if isinstance(oldItem, str):
                    toks = oldItem.split(":")
                    if len(toks) == 2 and (toks[1] == "=" or toks[0] == "NAME"):
                        raise GrpcApiError("The element is a key. It should not be overwritten.")
            return super().__setitem__(idxOrKey, newVal)
        if isinstance(idxOrKey, str):
            idx = self.__GetValueIdxByKey__(idxOrKey)
            if idx != None:
                return super().__setitem__(idx, newVal)
            raise GrpcApiError("Key is not found.")
        raise GrpcApiError("Must be a key name or index.")

    def keys(self):
        arr = []
        for i in range(0, len(self) - 1):
            if isinstance(self[i], str):
                toks = self[i].split(":")
                if len(toks) == 2 and toks[1] == "=":
                    arr.append(toks[0])
        return arr


exclude_list = ["GetAppDesktop", "GetProcessID", "GetGrpcServerPort"]


class AedtObjWrapper:
    def __init__(self, objID, listFuncs, AedtAPI=None):
        self.__dict__["objectID"] = objID  # avoid derive class overwrite __setattr__
        self.__dict__["__methodNames__"] = listFuncs
        self.dllapi = AedtAPI
        self.is_linux = True if os.name == "posix" else False

    # print(self.objectID)

    def __str__(self):
        return "Instance of an Aedt object:" + str(self.objectID)

    def __Invoke__(self, funcName, argv):
        if settings.enable_debug_grpc_api_logger:
            settings.logger.debug("{} {}".format(funcName, argv))
        try:
            if (settings.use_multi_desktop and funcName not in exclude_list) or funcName in inclusion_list:
                self.dllapi.recreate_application(True)
            ret = _retry_ntimes(
                settings.number_of_grpc_api_retries,
                self.dllapi.AedtAPI.InvokeAedtObjMethod,
                self.objectID,
                funcName,
                argv,
            )  # Call C function
            if ret and isinstance(ret, (AedtObjWrapper, AedtPropServer)):
                ret.AedtAPI = self.AedtAPI
            return ret
        except Exception:  # pragma: no cover
            raise GrpcApiError("Failed to execute gRPC AEDT command: {}".format(funcName))

    def __dir__(self):
        return self.__methodNames__

    def __GetObjMethod__(self, funcName):
        try:

            def DynamicFunc(self, *args):
                return self.__Invoke__(funcName, args)

            return types.MethodType(DynamicFunc, self)
        except (AttributeError, GrpcApiError):
            raise GrpcApiError("This AEDT object has no attribute '" + funcName + "'")

    def __getattr__(self, funcName):
        try:
            if funcName == "ScopeID":  # backward compatible for IronPython wrapper.
                return self.objectID
            return self.__GetObjMethod__(funcName)
        except Exception:
            raise GrpcApiError("Failed to get gRPC API AEDT attribute {}".format(funcName))

    def __setattr__(self, attrName, val):
        if attrName == "objectID" or attrName == "__methodNames__":
            raise GrpcApiError("This attribute cannot be modified.")
        elif attrName in self.__methodNames__:
            raise GrpcApiError(attrName + " is a function name.")
        else:
            super().__setattr__(attrName, val)

    def __del__(self):
        if "ReleaseAedtObject" in dir(self.dllapi):
            self.dllapi.ReleaseAedtObject(self.objectID)

    def match(self, patternStr):  # IronPython wrapper implemented this function return IEnumerable<string>.
        class IEnumerable(list):
            def __getattr__(self, key):
                if key == "Count":
                    return len(self)

        pattern = re.compile(patternStr)
        found = IEnumerable()
        allMethods = self.__methodNames__
        for method in allMethods:
            if pattern.match(method):
                found.append(method)
        return found

    def GetHashCode(self):  # IronPython build in function
        return self.__hash__()


class AedtPropServer(AedtObjWrapper):
    def __init__(self, objID, listFuncs, aedtapi):
        AedtObjWrapper.__init__(self, objID, listFuncs, aedtapi)
        self.__dict__["__propMap__"] = None
        self.__dict__["__propNames__"] = None

    def __GetPropAttributes(self):
        if self.__propMap__ == None:
            propMap = {}
            propNames = self.GetPropNames()
            for prop in propNames:
                attrName = ""
                if prop[0].isdigit():
                    attrName += "_"
                for c in prop:
                    if c.isalnum() == True:
                        attrName += c
                    else:
                        attrName += "_"
                propMap[attrName] = prop
            self.__propMap__ = propMap
        return self.__propMap__

    def __dir__(self):
        ret = super().__dir__().copy()
        for attrName, _ in self.__GetPropAttributes():
            ret.append(attrName)
        return ret

    def __getattr__(self, attrName):
        try:
            return super().__getattr__(attrName)
        except AttributeError:
            # if AedtAPI.IsAedtObjPropName(self.objectID, attrName, False):
            #    return self.GetPropValue(attrName)
            propMap = self.__GetPropAttributes()
            if attrName in propMap:
                return self.GetPropValue(propMap[attrName])
            raise GrpcApiError("Failed to retrieve attribute {} from gRPC API".format(attrName))

    def __setattr__(self, attrName, val):
        if attrName in self.__dict__:
            self.__dict__[attrName] = val
            return
        propMap = self.__GetPropAttributes()
        try:
            if attrName in propMap:
                self.SetPropValue(propMap[attrName], val)
                return
        except Exception:
            pass
        super().__setattr__(attrName, val)

    def GetName(self):
        return self.__Invoke__("GetName", ())

    def GetObjPath(self):
        return self.__Invoke__("GetObjPath", ())

    def GetChildNames(self, childType=""):
        return self.__Invoke__("GetChildNames", (childType))

    def GetPropNames(self, includeReadOnly=True):
        if includeReadOnly:
            if self.__propNames__ == None:
                self.__propNames__ = self.__Invoke__("GetPropNames", (includeReadOnly,))
            return self.__propNames__
        return self.__Invoke__("GetPropNames", (includeReadOnly,))

    def GetPropValue(self, propName=""):
        return self.__Invoke__("GetPropValue", (propName,))

    def SetPropValue(self, propName, val):
        return self.__Invoke__("SetPropValue", (propName, val))


class AEDT:
    def __init__(self, pathDir):
        is_linux = os.name == "posix"
        is_windows = not is_linux
        self.original_path = pathDir
        self.pathDir = pathDir
        self.pathDir = os.path.dirname(self.pathDir)  # PythonFiles
        self.pathDir = os.path.dirname(self.pathDir)  # DesktopPlugin or Win64
        # dirName = os.path.basename(pathDir)

        # Plugin filename depends on OS
        if is_linux:
            pluginFileName = r"libPyDesktopPlugin.so"
        else:
            pluginFileName = r"PyDesktopPlugin.dll"

        AedtAPIDll_file = os.path.join(self.pathDir, pluginFileName)  # install dir

        if not os.path.isfile(AedtAPIDll_file):
            self.pathDir = os.path.dirname(self.pathDir)  # lib
            self.pathDir = os.path.dirname(self.pathDir)  # core
            self.pathDir = os.path.dirname(self.pathDir)  # view
            AedtAPIDll_file = os.path.join(self.pathDir, r"build_output\64Release\PyDesktopPlugin.dll")  # develop dir
            # AedtAPIDll_file = os.path.join(pathDir, r"PyAedtStub/x64/Debug/PyAedtStub.dll") #develop dir

        # load dll
        if is_windows:
            # on windows, modify path
            aedtDir = os.path.dirname(AedtAPIDll_file)
            originalPath = os.environ["PATH"]
            os.environ["PATH"] = originalPath + os.pathsep + aedtDir
            AedtAPI = PyDLL(AedtAPIDll_file)
            os.environ["PATH"] = originalPath
        else:
            AedtAPI = PyDLL(AedtAPIDll_file)
        # AedtAPI.SetPyObjCalbacks.argtypes = py_object, py_object, py_object
        AedtAPI.SetPyObjCalbacks.restype = None

        # Must use global variable to hold those functions reference
        self.callbackToCreateObj = None
        self.callbackCreateBlock = None
        self.callbackGetObjID = None
        version = None
        with open(os.path.join(self.pathDir, "product.info"), "r") as f:
            for line in f:
                if "AnsProductVersion" in line:
                    version = line.split("=")[1].strip('\n"')
                    break

        if version >= "24.1":
            AedtAPI.CreateAedtApplication.argtypes = c_wchar_p, py_object, c_bool, c_bool
        else:
            AedtAPI.CreateAedtApplication.argtypes = c_wchar_p, c_int, c_bool, c_bool

        AedtAPI.CreateAedtApplication.restype = py_object

        AedtAPI.InvokeAedtObjMethod.argtypes = c_int, c_wchar_p, py_object
        AedtAPI.InvokeAedtObjMethod.restype = py_object

        AedtAPI.ReleaseAedtObject.argtypes = (c_int,)
        AedtAPI.ReleaseAedtObject.restype = None
        self.AedtAPI = AedtAPI
        self.SetPyObjCalbacks()
        self.aedt = None
        self.non_graphical = False

    def SetPyObjCalbacks(self):
        self.callback_type = CFUNCTYPE(py_object, c_int, c_bool, py_object)
        self.callbackToCreateObj = self.callback_type(
            self.CreateAedtObj
        )  # must use global variable to hold this function reference
        RetObj_InObj_Func_type = CFUNCTYPE(py_object, py_object)
        self.callbackCreateBlock = RetObj_InObj_Func_type(self.CreateAedtBlockObj)
        self.callbackGetObjID = RetObj_InObj_Func_type(self.GetAedtObjId)
        self.AedtAPI.SetPyObjCalbacks(self.callbackToCreateObj, self.callbackCreateBlock, self.callbackGetObjID)

    @staticmethod
    def _check_grpc_port(port, machine_name=""):
        s = socket.socket()
        try:
            if not machine_name:
                machine_name = "127.0.0.1"
            s.connect((machine_name, port))
        except socket.error:
            success = False
        else:
            success = True
        finally:
            s.close()
        return success

    def CreateAedtApplication(self, machine="", port=0, NGmode=False, alwaysNew=True):
        try:
            self.aedt = self.AedtAPI.CreateAedtApplication(machine, port, NGmode, alwaysNew)
        except Exception:  # pragma: no cover
            if port and self._check_grpc_port(port):
                time.sleep(5)  # waiting for Desktop to initialize Grpc Server
                self.aedt = self.AedtAPI.CreateAedtApplication(machine, port, NGmode, alwaysNew)
            if not self.aedt:
                raise GrpcApiError("Failed to connect to Desktop Session")
        self.machine = machine
        self.non_graphical = NGmode
        if port == 0:
            self.port = self.aedt.GetAppDesktop().GetGrpcServerPort()
        else:
            self.port = port

        return self.aedt

    @property
    def odesktop(self):
        if settings.use_multi_desktop:
            self.aedt = self.recreate_application()
        return self.aedt.GetAppDesktop()

    def recreate_application(self, force=False):
        def run():
            self.ReleaseAedtObject(self.aedt.objectID)
            port = self.port
            machine = self.machine
            self.__init__(self.original_path)
            self.port = port
            self.machine = machine
            self.aedt = self.AedtAPI.CreateAedtApplication(self.machine, self.port, self.non_graphical, False)
            return self.aedt

        if force:
            return run()
        else:
            try:
                self.aedt.GetAppDesktop()
                return self.aedt
            except Exception:
                return run()

    def InvokeAedtObjMethod(self, objectID, funcName, argv):
        return self.AedtAPI.InvokeAedtObjMethod(objectID, funcName, argv)

    def ReleaseAedtObject(self, objectID):
        self.AedtAPI.ReleaseAedtObject(objectID)

    def ReleaseAll(self):
        self.AedtAPI.ReleaseAll()

    def IsEmbedded(self):
        return False

    def CreateAedtObj(self, objectID, bIsPropSvr, listFuncs):
        # print("Create " + str(objectID))
        if bIsPropSvr:
            return AedtPropServer(
                objectID,
                listFuncs,
                self,
            )

        return AedtObjWrapper(
            objectID,
            listFuncs,
            self,
        )

    def CreateAedtBlockObj(self, list_in):
        count = len(list_in)
        if count > 1:
            if isinstance(list_in[0], str):
                toks = list_in[0].split(":")
                if len(toks) == 2:
                    start = -1
                    if count % 2 == 0:
                        if toks[1] == "=":
                            start = 2
                    elif count > 2:
                        if toks[0] == "NAME":
                            start = 1
                    if start > 0:
                        isBlock = True
                        for i in range(start, count - 1, 2):
                            if isinstance(list_in[i], str):
                                toks = list_in[i].split(":")
                                if len(toks) != 2 or toks[1] != "=":
                                    isBlock = False
                                    break
                            else:
                                isBlock = False
                                break
                        if isBlock:
                            return AedtBlockObj(list_in)
        return list_in

    def GetAedtObjId(self, obj):
        if isinstance(obj, AedtObjWrapper):
            return obj.objectID
        return None

    def Release(self):
        self.AedtAPI.ReleaseAll()
